@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker and Richard O'Keefe
     E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2014-2019, VU University Amsterdam
+    Copyright (c)  2014-2023, VU University Amsterdam
                               CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -41,13 +42,14 @@
             test_installation/0,
             test_installation/1                 % +Options
           ]).
-:- autoload(library(apply),[maplist/2,maplist/3]).
-:- autoload(library(archive),[archive_open/3,archive_close/1]).
-:- autoload(library(lists),[append/3,member/2]).
-:- autoload(library(option),[option/2,merge_options/3]).
-:- autoload(library(pcre),[re_config/1]).
-:- autoload(library(prolog_source),[path_segments_atom/2]).
-:- use_module(library(settings),[setting/2]).
+:- autoload(library(apply), [maplist/2, maplist/3]).
+:- autoload(library(archive), [archive_open/3, archive_close/1]).
+:- autoload(library(lists), [append/3, member/2]).
+:- autoload(library(option), [option/2, merge_options/3]).
+:- autoload(library(prolog_source), [path_segments_atom/2]).
+:- use_module(library(settings), [setting/2]).
+:- autoload(library(dcg/high_order), [sequence//2, sequence/4]).
+:- autoload(library(error), [must_be/2]).
 
 
 /** <module> Check installation issues and features
@@ -84,7 +86,8 @@ shared objects/DLLs can be loaded.
 component(tcmalloc,
           _{ optional:true,
              test:test_tcmalloc,
-             url:'tcmalloc.html'
+             url:'tcmalloc.html',
+             os:linux
            }).
 component(gmp,
           _{ test:current_prolog_flag(bounded, false),
@@ -124,6 +127,7 @@ component(library(sha), _{}).
 component(library(snowball), _{}).
 component(library(socket), _{}).
 component(library(ssl), _{}).
+component(library(sweep_link), _{features:sweep_emacs_module}).
 component(library(crypto), _{}).
 component(library(syslog), _{os:unix}).
 component(library(table), _{}).
@@ -134,6 +138,7 @@ component(library(uri), _{}).
 component(library(uuid), _{}).
 component(library(zlib), _{}).
 component(library(yaml), _{}).
+component(library(janus), _{features:python_version}).
 
 issue_base('http://www.swi-prolog.org/build/issues/').
 
@@ -199,6 +204,10 @@ check_component(Source) :-
     !,
     check_component(Source, Properties.put(source,Source)).
 
+check_component(_Source, Properties) :-
+    OS = Properties.get(os),
+    \+ current_os(OS),
+    !.
 check_component(Source, Properties) :-
     compound(Source),
     !,
@@ -210,10 +219,6 @@ check_component(Feature, Properties) :-
     ;   print_issue(installation(missing(Properties)))
     ).
 
-check_source(_Source, Properties) :-
-    OS = Properties.get(os),
-    \+ current_os(OS),
-    !.
 check_source(Source, Properties) :-
     exists_source(Source),
     !,
@@ -419,6 +424,26 @@ jquery_file :-
     ;   print_message(warning, installation(jquery(not_found(File))))
     ).
 
+sweep_emacs_module :-
+    with_output_to(string(S), write_sweep_module_location),
+    split_string(S, "\n", "\n", [VersionInfo|Modules]),
+    must_be(oneof(["V 1"]), VersionInfo),
+    (   maplist(check_sweep_lib, Modules)
+    ->  print_message(informational, installation(sweep(found(Modules))))
+    ;   print_message(warning, installation(sweep(not_found(Modules))))
+    ).
+
+check_sweep_lib(Line) :-
+    sub_atom(Line, B, _, A, ' '),
+    sub_atom(Line, 0, B, _, Type),
+    must_be(oneof(['L', 'M']), Type),
+    sub_atom(Line, _, A, 0, Lib),
+    exists_file(Lib).
+
+python_version :-
+    py_call(sys:version, Version),
+    print_message(informational, installation(janus(Version))).
+
 
 %!  check_on_path
 %
@@ -547,7 +572,7 @@ message(checking) -->
     [ 'Installed at: ~`.t~24| ~w'-[Home] ], [nl],
     [ 'Cores: ~`.t~24| ~w'-[Cores] ], [nl],
     [ nl ].
-message(perfect) -->
+>message(perfect) -->
     [ nl, 'Congratulations, your kit seems sound and complete!'-[] ].
 message(imperfect(N)) -->
     [ 'Found ~w issues.'-[N] ].
@@ -607,10 +632,17 @@ message(jquery(found(Path))) -->
     [ '  jQuery from ~w'-[Path] ].
 message(jquery(not_found(File))) -->
     [ '  Cannot find jQuery (~w)'-[File] ].
+message(sweep(found(Paths))) -->
+    [ '  GNU-Emacs plugin loads'-[] ],
+    sequence(list_file, Paths).
+message(sweep(not_found(Paths))) -->
+    [ '  Could not find all GNU-Emacs libraries'-[] ],
+    sequence(list_file, Paths).
 message(testing(no_installed_tests)) -->
     [ '  Runtime testing is not enabled.', nl],
     [ '  Please recompile the system with INSTALL_TESTS enabled.' ].
-
+message(janus(Version)) -->
+    [ '  Python version ~w'-[Version] ].
 
 public_executable(EXE, PublicProg) :-
     file_base_name(EXE, Prog),
@@ -675,6 +707,9 @@ list_names([H|T]) -->
         list_names(T)
     ).
 
+list_file(File) -->
+    [ nl, '    '-[], url(File) ].
+
 
 		 /*******************************
 		 *          CONFIG FILES	*
