@@ -43,6 +43,7 @@
 		     [ type(oneof([text,binary])),
 		       alias(atom),
 		       encoding(encoding),
+                       newline(oneof([dos,posix,detect])),
 		       bom(boolean),
 		       create(list(atom)),
 		       eof_action(oneof([eof_code,error,reset])),
@@ -221,3 +222,12 @@
                      [ as(oneof([first,last])),
                        name(atom)
                      ]).
+:- predicate_options(system:open_shared_object/3, 3,
+                     [ resolve(oneof([lazy,now])),
+                       visibility(oneof([local,global])),
+                       now(bool),
+                       global(bool),
+		       delete(bool),
+		       load(bool),
+		       deepbind(bool)
+                     ]).
