@@ -57,6 +57,7 @@ particularly integer conversions.
 #define _SWI_CPP2_H
 
 #include <SWI-Prolog.h>
+#include <SWI-Stream.h>
 #include <climits>
 #include <cstdint>
 #include <cstring>
@@ -64,6 +65,7 @@ particularly integer conversions.
 #include <functional>
 #include <string>
 #include <cassert>
+#include <memory>
 
 #if INT_MAX != 0x7fffffff
   #error "Unexpected value for INT_MAX"
@@ -91,6 +93,15 @@ class PlTerm;
 class PlTermv;
 class PlRecord;
 class PlRecordExternalCopy;
+class PlBlob;
+
+
+// PlExceptionBase is used for try-catch that handles the
+// exceptions defined in this header file but excluds the
+// standard C++ exceptions.
+class PlExceptionBase : public std::exception
+{
+};
 
 
 // PlFail is a pseudo-exception for quick exist on failure, for use by
@@ -99,7 +110,7 @@ class PlRecordExternalCopy;
 // a subclass of PlException. See the documentation for more details
 // on how this works with returning Prolog failure and returning
 // exceptions.
-class PlFail : public std::exception
+class PlFail : public PlExceptionBase
 {
 public:
   explicit PlFail() {}
@@ -112,7 +123,7 @@ public:
 // PlExceptionFail is a variant of PlFail, for when a resource error
 // happens and we can't use PlException (because we're out of resources
 // and therefore can't create any more terms).
-class PlExceptionFail : public std::exception
+class PlExceptionFail : public PlExceptionBase
 {
 public:
   explicit PlExceptionFail() {}
@@ -129,7 +140,7 @@ public:
 // wish to wrap the call in PlCheckFail().
 template<typename C_t> [[nodiscard]] C_t PlWrap(C_t rc, qid_t qid = 0);
 
-// As PlWrap, but always throw an exception for non-zero rc.
+// As PlWrap, but always throw an exception for zero rc.
 // This is for functions that report errors but don't have an
 // indication of "fail" - that is, almost everything except for
 // functions like PL_unify_*() or PL_next_solution().
@@ -249,6 +260,8 @@ public:
   }
   const std::wstring as_wstring() const;
 
+  [[nodiscard]] int write(IOSTREAM *s, int flags) const;
+
   // TODO: operator == should be `override`
   bool operator ==(const PlAtom& to) const /*override*/ { return C_ == to.C_; }
   bool operator !=(const PlAtom& to) const /*override*/ { return C_ != to.C_; }
@@ -262,6 +275,8 @@ public:
   bool operator !=(const wchar_t *s) const { return !eq(s); }
   [[deprecated("use PlAtom instead of atom_t")]] bool operator !=(atom_t to) const { return C_ != to; }
 
+  // TODO: when C++17 becomes standard, rename register_ref() to register().
+
   void register_ref() const
   { Plx_register_atom(C_);
   }
@@ -585,7 +600,12 @@ public:
   [[nodiscard]] bool unify_nil()                         const { return Plx_unify_nil(C_); }
   [[nodiscard]] bool unify_list(PlTerm h, PlTerm t)      const { return Plx_unify_list(C_, h.C_, t.C_); }
   [[nodiscard]] bool unify_bool(bool val)                const { return Plx_unify_bool(C_, val); }
-  [[nodiscard]] bool unify_blob(void *blob, size_t len, PL_blob_t *type) const { return Plx_unify_blob(C_, blob, len, type); }
+
+
+  [[nodiscard]] bool unify_blob(const PlBlob* blob) const;
+  [[nodiscard]] bool unify_blob(std::unique_ptr<PlBlob>* blob) const;
+  [[nodiscard]] bool unify_blob(const void *blob, size_t len, const PL_blob_t *type) const
+  { return Plx_unify_blob(C_, const_cast<void*>(blob), len, const_cast<PL_blob_t*>(type)); }
 
   // TODO: PL_unify_mpz(), PL_unify_mpq()
 
@@ -620,8 +640,11 @@ public:
   bool operator !=(const std::wstring& s) const { return !eq(s); }
   bool operator !=(const PlAtom& a)       const { return !eq(a); }
 
-  // E.g.: t.write(Serror, 1200, PL_WRT_NEWLINE|PL_WRT_QUOTED);
-  void write(IOSTREAM *s, int precedence, int flags) const { Plx_write_term(s, C_, precedence, flags); }
+  // E.g.: t.write(Serror, 1200, PL_WRT_QUOTED);
+  [[nodiscard]] int write(IOSTREAM *s, int precedence, int flags) const
+  { // TODO: move "&~PL_WRITE_NEWLINE" to PL_write_term() and update foreign.doc
+    return Plx_write_term(s, C_, precedence, flags & ~PL_WRT_NEWLINE);
+  }
 
   void reset_term_refs() { Plx_reset_term_refs(C_); }
 
@@ -652,6 +675,13 @@ public:
   explicit PlTerm_atom(const std::wstring& text) { PlEx<int>(Plx_unify_wchars(C_, PL_ATOM, text.size(), text.data())); }
 };
 
+[[nodiscard]]
+inline int
+PlAtom::write(IOSTREAM *s, int flags) const {
+  // TODO: move "&~PL_WRITE_NEWLINE" to PL_write_term() and update foreign.doc
+  return PlTerm_atom(*this).write(s, 1200, flags & ~PL_WRT_NEWLINE);
+}
+
 class PlTerm_var : public PlTerm
 {
 public:
@@ -848,13 +878,13 @@ public:
   }
 
   void erase()
-  { if ( C_ != null )
+  { if ( not_null() )
       Plx_erase(C_);
-    C_ = null;
+    reset();
   }
 
   PlRecord duplicate() const
-  { return PlRecord(Plx_duplicate_record(C_));
+  { return is_null() ? PlRecord(null) : PlRecord(Plx_duplicate_record(C_));
   }
 
   ~PlRecord()
@@ -925,7 +955,7 @@ private:
 //       - the same as below
 //       - PlExceptionFailImpl - for error(resource_error(_))
 
-class PlException : public std::exception
+class PlException : public PlExceptionBase
 {
 public:
   explicit PlException(const PlTerm& t)
@@ -974,13 +1004,14 @@ public:
     // allocating the std::string) even though we specify "throw()" -
     // telling the truth "noexcept(false)" results in a compilation
     // error.
+    (void)enc; // TODO: use this
     const_cast<PlException*>(this)->set_what_str();
     return what_str_;
   }
 
   // plThrow() is for the try-catch in PREDICATE - returns the result
   // of Plx_raise_exception(), which is always `false`, as a foreign_t.
-  virtual foreign_t plThrow()
+  virtual foreign_t plThrow() const
   { foreign_t rc = static_cast<foreign_t>(Plx_raise_exception(term().C_));
     return rc;
   }
@@ -1036,21 +1067,21 @@ PlException PlResourceError(const char *resource);
 
 PlException PlUnknownError(const char *description);
 
-void PlWrap_impl(qid_t qid);
+void PlWrap_fail(qid_t qid);
 
 template<typename C_t> C_t
 PlWrap(C_t rc, qid_t qid)
 { if ( rc == static_cast<C_t>(0) )
-    PlWrap_impl(qid);
+    PlWrap_fail(qid);
   return rc;
 }
 
-void PlEx_impl(qid_t qid);
+void PlEx_fail(qid_t qid);
 
 template<typename C_t> void
 PlEx(C_t rc, qid_t qid)
 { if ( rc == static_cast<C_t>(0) )
-    PlEx_impl(qid);
+    PlEx_fail(qid);
 }
 
 
@@ -1099,6 +1130,14 @@ public:
   [[nodiscard]] void *foreign_context_address() const { return Plx_foreign_context_address(C_); }
 
   [[nodiscard]] PlPredicate foreign_context_predicate() const { return PlPredicate(Plx_foreign_context_predicate(C_)); }
+
+  template <typename ContextType>
+  [[nodiscard]]
+  std::unique_ptr<ContextType>
+  context_unique_ptr()
+  { return std::unique_ptr<ContextType>(static_cast<ContextType *>(foreign_context_address()));
+  }
+
 };
 
 
@@ -1207,6 +1246,7 @@ public:
   // The return code from next_solution can be (if called with PL_Q_EXT_STATUS):
   //    TRUE
   //    FALSE
+  //    PL_S_NOT_INNER
   //    PL_S_EXCEPTION
   //    PL_S_FALSE
   //    PL_S_TRUE
@@ -1363,24 +1403,34 @@ public:
 #define PROLOG_MODULE static_cast<const char*>(nullptr)
 #endif
 
+// This macro is used by both the PREDICATE macros and the blob callbacks
+
+#define PREDICATE_CATCH(error_action) \
+	    catch ( const std::bad_alloc& ) \
+	  { (void)Plx_resource_error("memory"); \
+            error_action; \
+	  } catch ( const PlFail& ) \
+	  { error_action; \
+	  } catch ( const PlExceptionFail& ) \
+          { error_action; \
+	  } catch ( const PlException& ex ) \
+	  { ex.plThrow(); \
+            error_action; \
+	  }
+
 #define NAMED_PREDICATE(plname, name, arity) \
 	static foreign_t \
 	pl_ ## name ## __ ## arity(PlTermv PL_av); \
 	static foreign_t \
 	_pl_ ## name ## __ ## arity(term_t t0, int a, control_t c) \
 	{ (void)a; (void)c; \
+          foreign_t rc; \
 	  try \
 	  { \
-	    return pl_ ## name ## __ ## arity(PlTermv(arity, PlTerm_term_t(t0))); \
-	  } catch ( const std::bad_alloc& ) \
-	  { return static_cast<foreign_t>(Plx_resource_error("memory")); \
-	  } catch ( const PlFail& ) \
-          { return false;     \
-	  } catch ( const PlExceptionFail& ) \
-          { return false;     \
-	  } catch ( PlException& ex ) \
-          { return ex.plThrow();  \
+	    rc = pl_ ## name ## __ ## arity(PlTermv(arity, PlTerm_term_t(t0))); \
 	  } \
+          PREDICATE_CATCH(rc = false) \
+          return rc; \
 	} \
 	static PlRegister _x ## name ## __ ## arity(PROLOG_MODULE, plname, arity, \
 					    _pl_ ## name ## __ ## arity); \
@@ -1392,18 +1442,13 @@ public:
 	static foreign_t \
 	_pl_ ## name ## __0(term_t t0, int a, control_t c) \
 	{ (void)t0; (void)a; (void)c; \
+          foreign_t rc; \
 	  try \
 	  { \
-	    return pl_ ## name ## __0(); \
-	  } catch ( std::bad_alloc& ) \
-	  { return static_cast<foreign_t>(Plx_resource_error("memory")); \
-	  } catch ( PlFail& ) \
-	  { return false; \
-	  } catch ( const PlExceptionFail& ) \
-          { return false;     \
-	  } catch ( PlException& ex ) \
-	  { return ex.plThrow(); \
+	    rc = pl_ ## name ## __0(); \
 	  } \
+          PREDICATE_CATCH(rc = false) \
+          return rc; \
 	} \
 	static PlRegister _x ## name ## __0(PROLOG_MODULE, plname, 0, \
 					    _pl_ ## name ## __0); \
@@ -1415,19 +1460,14 @@ public:
 	static foreign_t \
 	_pl_ ## name ## __ ## arity(term_t t0, int a, control_t c) \
 	{ (void)a; \
+          foreign_t rc; \
 	  try \
 	  { \
 	    /* t0.C_ is 0 if handle.foreign_control()==PL_PRUNED */ \
-	    return pl_ ## name ## __ ## arity(PlTermv(arity, PlTerm_term_t(t0)), PlControl(c)); \
-	  } catch ( std::bad_alloc& ) \
-	  { return static_cast<foreign_t>(Plx_resource_error("memory")); \
-	  } catch ( PlFail& ) \
-	  { return false; \
-	  } catch ( const PlExceptionFail& ) \
-          { return false;     \
-	  } catch ( PlException& ex ) \
-	  { return ex.plThrow(); \
+	    rc = pl_ ## name ## __ ## arity(PlTermv(arity, PlTerm_term_t(t0)), PlControl(c)); \
 	  } \
+          PREDICATE_CATCH(rc = false) \
+          return rc; \
 	} \
 	static PlRegister _x ## name ## __ ## arity(PROLOG_MODULE, plname, arity, \
 						    _pl_ ## name ## __ ## arity, \
@@ -1462,55 +1502,239 @@ public:
 #define A10	PL_A10
 #endif
 
+// TODO: delete this section
 		 /*******************************
 		 *     NONDET HELPERS		*
 		 *******************************/
 
-// For non-deterministic predicates that allocate a context, the
-// PlForeignContextPtr is a RAII (Resource Acquisition Is
-// Initialization) class (a kind of "smart pointer") that holds a
-// pointer that is deleted at exit (whether by return or exception)
-// unless deferred_free_ is false. This simplifies code that can throw
-// exceptions such as the PlTerm::as_long().
-//
-// The pointer must have been allocated using the `new` operator.  The
-// pointer can have the value nullptr.
-//
-// THe methods are the usual "smart pointer" ones: dereference (either
-// using `->` or `*`), get(), set(ptr). In addition:
-//    deferred_free() - the pointer will be deleted on return/throw
-//    keep() - the pointer will not be deleted on return/throw
-
-template <typename ContextType> class PlForeignContextPtr
+template <typename ContextType>
+// TODO: [[deprecated("Use PlControl::context_unique_ptr")]]
+class PlForeignContextPtr
 {
 public:
+  [[deprecated("Use PlControl::context_unique_ptr")]]
   explicit PlForeignContextPtr<ContextType>(PlControl handle)
-    : ptr_(static_cast<ContextType *>(handle.foreign_context_address())),
-      deferred_free_(true)
-  { }
+  { ptr_.reset(static_cast<ContextType *>(handle.foreign_context_address()));
+  }
 
   PlForeignContextPtr<ContextType>(const PlForeignContextPtr<ContextType>&) = delete;
   PlForeignContextPtr<ContextType>(PlForeignContextPtr<ContextType>&&) = delete;
   PlForeignContextPtr<ContextType>& operator =(const PlForeignContextPtr<ContextType>&) = delete;
   PlForeignContextPtr<ContextType>& operator =(PlForeignContextPtr<ContextType>&&) = delete;
 
-  ContextType& operator*()  const { return *ptr_; }
-  ContextType* operator->() const { return ptr_; }
-  ContextType* get()        const { return ptr_; }
-  void set(ContextType* ptr = nullptr) { ptr_ = ptr; }
-
-  void deferred_free()       { deferred_free_ = true; }
-  void deferred_free(bool v) { deferred_free_ = v; }
-  void keep()                { deferred_free_ = false; }
-
-  ~PlForeignContextPtr()
-  { if ( deferred_free_ )
-      delete ptr_; // it's safe to delete nullptr
-  }
+  [[deprecated("Use PlControl::context_unique_ptr")]] ContextType& operator*()  const { return *ptr_; }
+  [[deprecated("Use PlControl::context_unique_ptr")]] ContextType* operator->() const { return ptr_.get(); }
+  [[deprecated("Use PlControl::context_unique_ptr")]] ContextType* get()        const { return ptr_.get(); }
+  [[deprecated("Use PlControl::context_unique_ptr")]] void set(ContextType* ptr = nullptr)   { reset(ptr); } // TODO: delete
+  [[deprecated("Use PlControl::context_unique_ptr")]] void reset(ContextType* ptr = nullptr) { ptr_.reset(ptr); }
+  [[deprecated("Use PlControl::context_unique_ptr")]] ContextType* keep()        { return release(); } // TODO: delete
+  [[deprecated("Use PlControl::context_unique_ptr")]] ContextType* release()     { return ptr_.release(); }
 
 private:
-  ContextType *ptr_;
-  bool deferred_free_;
+  std::unique_ptr<ContextType> ptr_;
 };
 
+
+		 /*******************************
+		 *          BLOBS		*
+		 *******************************/
+
+// PlBlobV is used by PL_BLOB_DEFINITION to create a vector of
+// C-callable functions that in turn call the methods inside PlBlob.
+// Although PlBlobV is defined as a class, it is really a namespace,
+// with a template variable C_t for convenience.
+//
+// TODO: This code assumes that the only difference between C and C++
+//       functions is the name mangling. If there are additional
+//       differences, this template class will need to be replaced by
+//       some macros that define a set of `extern "C"` functions.
+
+template<typename C_t>
+class PlBlobV
+{
+public:
+  [[nodiscard]]
+  static C_t *
+  cast(PlAtom aref)
+  { size_t len;
+    PL_blob_t *type;
+    auto ref = static_cast<C_t *>(aref.blob_data(&len, &type));
+    // Can't throw PlException here because might be in a context
+    // outside of a PREDICATE.
+    if ( ref && type == ref->blob_t_ )
+    { assert(len == sizeof *ref);
+      return ref;
+    }
+    return nullptr;
+  }
+
+  [[nodiscard]]
+  static C_t*
+  cast_check(PlAtom aref)
+  { auto ref = cast(aref);
+    // Can't throw PlException here because might be in a context
+    // outside of a PREDICATE.
+    assert(ref);
+    return ref;
+  }
+
+  [[nodiscard]]
+  static C_t*
+  cast_ex(PlTerm t, const PL_blob_t& b)
+  { auto ref = cast(t.as_atom());
+    if ( !ref )
+      throw PlTypeError(b.name, t);
+    return ref;
+  }
+
+  static void acquire(atom_t a)
+  { PlAtom a_(a);
+    auto data = cast_check(a_);
+    bool rc;
+    try
+    { data->acquire(a_);
+      rc = true;
+    }
+    PREDICATE_CATCH(rc = false)
+      // TODO: if ( ! rc ) Plx_clear_exception() ?
+    assert(rc);
+    (void)rc;
+  }
+
+  [[nodiscard]]
+  static int release(atom_t a) noexcept
+  { auto data = cast_check(PlAtom(a));
+    delete data;
+    return true;
+  }
+
+  [[nodiscard]]
+  static int compare(atom_t a, atom_t b)
+  { if (a == b) // Just in case Prolog didn't check this.
+      return 0;
+    // Prolog should never give us two atoms (blobs) with different
+    // types - they should have been already compared by standard
+    // order of types; but use cast_check() anyway (which will be
+    // optimised away if NDEBUG).
+    bool rc_try;
+    int rc = -1; // Stop the compiler warning about uninitialized
+    try
+    { const auto a_data = cast_check(PlAtom(a)); // TODO: cast(PlAtom(a))
+      const auto b_data = cast_check(PlAtom(b));
+      rc = a_data->compare_fields(b_data);
+      if ( rc == 0 )
+      { rc = (a_data < b_data) ? -1 : (a_data > b_data) ? 1 : 0;
+      }
+      rc_try = true;
+    }
+    PREDICATE_CATCH(rc_try = false)
+      // TODO: if ( ! rc_try ) Plx_clear_exception() ?
+    assert(rc_try);
+    (void)rc_try;
+    return rc;
+  }
+
+  [[nodiscard]]
+  static int write(IOSTREAM *s, atom_t a, int flags)
+  { const auto data = cast_check(PlAtom(a));
+    int rc;
+    try
+    { rc = data->write(s, flags);
+    }
+    PREDICATE_CATCH(rc = -1)
+      // TODO: if ( rc < 0 ) Plx_clear_exception() ?
+    return rc;
+  }
+
+  [[nodiscard]]
+  static int save(atom_t a, IOSTREAM *fd)
+  { const auto data = cast_check(PlAtom(a));
+    bool rc;
+    try
+    { data->save(fd);
+      rc = true;
+    }
+    PREDICATE_CATCH(rc = false)
+      // TODO: if ( ! rc ) Plx_clear_exception() ?
+    return rc;
+  }
+
+  [[nodiscard]]
+  static atom_t load(IOSTREAM *fd)
+  { C_t ref;
+    atom_t atom;
+    int rc_try;
+    try
+    { atom = ref.load(fd).C_;
+      rc_try = true;
+    }
+    PREDICATE_CATCH(rc_try = false);
+    if ( !rc_try )
+    { // TODO: Plx_clear_exception() ?
+      return PlAtom::null;
+    }
+    return atom;
+  }
+};
+
+
+#define PL_BLOB_DEFINITION(blob_class, blob_name) \
+{ .magic   = PL_BLOB_MAGIC,			\
+  .flags   = PL_BLOB_NOCOPY,			\
+  .name    = blob_name,				\
+  .release = PlBlobV<blob_class>::release,	\
+  .compare = PlBlobV<blob_class>::compare,	\
+  .write   = PlBlobV<blob_class>::write,	\
+  .acquire = PlBlobV<blob_class>::acquire,	\
+  .save    = PlBlobV<blob_class>::save,		\
+  .load    = PlBlobV<blob_class>::load		\
+}
+
+#define PL_BLOB_SIZE \
+  virtual size_t blob_size_() const override { return sizeof *this; }
+
+class PlBlob
+{
+public:
+  explicit PlBlob(const PL_blob_t* _blob_t)
+    : blob_t_(_blob_t) { }
+  explicit PlBlob() = delete;
+  explicit PlBlob(const PlBlob&) = delete;
+  explicit PlBlob(PlBlob&&) = delete;
+  PlBlob& operator =(const PlBlob&) = delete;
+  virtual ~PlBlob() = default;
+
+  virtual size_t blob_size_() const = 0; // See PL_BLOB_SIZE
+
+  void acquire(PlAtom _symbol)
+  { symbol_ = _symbol;
+    // Don't: symbol_.register_ref() because it's already got a
+    // reference from when the blob was allocated, (by
+    // PL_unify_blob(), PL_put_blob(), or PL_new_blob()); extra
+    // reference count would prevent its GC.
+  }
+
+  PlTerm symbol_term() const;
+
+  virtual int compare_fields(const PlBlob *_b) const
+  { return 0; // compare() will do bitwise comparison
+  }
+
+  bool write(IOSTREAM *s, int flags) const;
+
+  bool virtual write_fields(IOSTREAM *s, int flags) const
+  { return true; }
+
+  virtual void save(IOSTREAM *fd) const;
+
+  virtual PlAtom load(IOSTREAM *fd);
+
+  const PL_blob_t* blob_t_ = nullptr;
+
+  // Associated symbol (used for error terms) filed in by acquire()
+  // and usually accessed by the symbol_term() method
+  PlAtom symbol_ = PlAtom(PlAtom::null);
+};
+
+
 #endif /*_SWI_CPP2_H*/
