@@ -43,7 +43,7 @@
             sweep_predicate_location/2,
             sweep_predicate_apropos/2,
             sweep_predicates_collection/2,
-            sweep_functor_arity_pi/2,
+            sweep_module_functor_arity_pi/2,
             sweep_modules_collection/2,
             sweep_packs_collection/2,
             sweep_pack_install/2,
@@ -63,9 +63,8 @@
             sweep_xref_source/2,
             sweep_beginning_of_next_predicate/2,
             sweep_beginning_of_last_predicate/2,
-            sweep_atom_collection/2,
             sweep_context_callable/2,
-            sweep_predicate_completion_candidates/2,
+            sweep_heads_collection/2,
             sweep_exportable_predicates/2,
             sweep_interrupt/0,
             sweep_string_to_atom/2,
@@ -86,7 +85,15 @@
             sweep_current_breakpoints/2,
             sweep_current_breakpoints_in_region/2,
             sweep_breakpoint_range/2,
-            sweep_breakpoint_file/2
+            sweep_breakpoint_file/2,
+            sweep_expand_macro/2,
+            sweep_module_annotation/2,
+            sweep_is_module/2,
+            sweep_module_class/2,
+            sweep_variable_start_code/2,
+            sweep_head_functors_collection/2,
+            sweep_functors_collection/2,
+            sweep_compound_functors_collection/2
           ]).
 
 :- use_module(library(pldoc)).
@@ -105,6 +112,10 @@
 :- use_module(library(prolog_deps)).
 :- use_module(library(dcg/high_order)).
 
+:- if(exists_source(library(macros))).
+:- use_module(library(macros)).
+:- endif.
+
 :- if(exists_source(library(help))).
 :- use_module(library(help)).
 :- endif.
@@ -168,7 +179,7 @@ sweep_current_prolog_flags(Sub, Flags) :-
              term_string(Value0, Value)),
             Flags).
 
-sweep_set_prolog_flag([Flag0|Value0], []) :-
+sweep_set_prolog_flag([Flag0|Value0], true) :-
     atom_string(Flag, Flag0),
     term_string(Value, Value0),
     set_prolog_flag(Flag, Value).
@@ -255,6 +266,10 @@ sweep_documentation_modes([mode(Mode0, Args)|_], OneLiner, Docs) :-
 sweep_documentation_modes([_|T], OneLiner, Docs) :-
     sweep_documentation_modes(T, OneLiner, Docs).
 
+sweep_module_class(M0, C) :-
+    atom_string(M, M0),
+    module_property(M, class(C0)),
+    atom_string(C0, C).
 
 sweep_module_path(ModuleName, Path) :-
     atom_string(Module, ModuleName),
@@ -267,7 +282,9 @@ sweep_module_path_(Module, Path) :-
 sweep_module_path_(Module, Path) :-
     xref_module(Path, Module), !.
 sweep_module_path_(Module, Path) :-
-    '$autoload':library_index(_, Module, Path0), atom_concat(Path0, '.pl', Path).
+    '$autoload':library_index(_, Module, Path0), atom_concat(Path0, '.pl', Path), !.
+sweep_module_path_(user, _).
+
 
 sweep_predicate_properties(P0, Props) :-
     term_string(P, P0),
@@ -276,9 +293,9 @@ sweep_predicate_properties(P0, Props) :-
 
 sweep_predicate_html_documentation(P0, D) :-
     term_string(P1, P0),
-    (   P1 = M:F/N
+    (   P1 = M:PI
     ->  true
-    ;   P1 = F/N, M = system
+    ;   P1 = PI, M = system
     ),
     (   (   current_module(M)
         ;   xref_module(_, M)
@@ -287,10 +304,10 @@ sweep_predicate_html_documentation(P0, D) :-
     ;   '$autoload':library_index(_, M, Path),
         xref_source(Path, [comments(store)])
     ),
-    (   M == system
-    ->  pldoc_man:load_man_object(F/N, _, _, DOM)
-    ;   doc_comment(M:F/N, Pos, _, Comment),
-        pldoc_html:pred_dom(M:F/N, [], Pos-Comment, DOM)
+    (   pldoc_man:load_man_object(M:PI, _, _, DOM)
+    ;   pldoc_man:load_man_object(PI, _, _, DOM)
+    ;   doc_comment(M:PI, Pos, _, Comment),
+        pldoc_html:pred_dom(M:PI, [], Pos-Comment, DOM)
     ),
     phrase(pldoc_html:html(DOM), HTML),
     with_output_to(string(D), html_write:print_html(HTML)).
@@ -309,36 +326,73 @@ sweep_module_html_documentation(M0, D) :-
     phrase(pldoc_html:html(DOM), HTML),
     with_output_to(string(D), html_write:print_html(HTML)).
 
-sweep_modules_collection([], Modules) :-
-    findall([M|P], ( module_property(M, file(P0)), atom_string(P0, P) ), Modules0, Tail),
-    setof([M|P], P0^N^('$autoload':library_index(N, M, P0), string_concat(P0, ".pl", P) ), Tail),
-    list_to_set(Modules0, Modules1),
-    maplist(sweep_module_description, Modules1, Modules).
+sweep_modules_collection([Bef|Aft], Ms) :-
+    setof(M, sweep_known_module(M), Ms0),
+    include(sweep_matching_module(Bef,Aft), Ms0, Ms1),
+    maplist(atom_string, Ms1, Ms).
 
-sweep_module_description([M0|P], [M|[P|D]]) :-
-   doc_comment(M0:module(D0), _, _, _),
-   atom_string(M0, M),
-   atom_string(D0, D).
-sweep_module_description([M0|P], [M|[P]]) :- atom_string(M0, M).
+sweep_matching_module([], Aft, Mod) :-
+    !,
+    sweep_matching_module_(Aft, 0, Mod).
+sweep_matching_module(Bef, Aft, Mod) :-
+    once(sub_atom(Mod, N, L, _, Bef)),
+    M is N + L,
+    sweep_matching_module_(Aft, M, Mod).
+
+sweep_matching_module_([], _, _) :- !.
+sweep_matching_module_(A, N, M) :-
+    sub_atom(M, B, _, _, A),
+    B >= N,
+    !.
+
+sweep_module_annotation(M0, [P|D]) :-
+    atom_string(M, M0),
+    (   sweep_module_path_(M, P0), nonvar(P0)
+    ->  sweep_file_path_in_library(P0, P)
+    ;   P = []
+    ),
+    (   sweep_module_description_(M, P0, D0)
+    ->  atom_string(D0, D)
+    ;   D = []
+    ).
+
+sweep_known_module(M) :-
+    current_module(M).
+sweep_known_module(M) :-
+    xref_module(_, M).
+sweep_known_module(M) :-
+    '$autoload':library_index(_, M, _).
+
+sweep_is_module(M0, _) :-
+    atom_string(M, M0),
+    once(sweep_known_module(M)).
+
+sweep_module_description_(M, _, D) :-
+    doc_comment(M:module(D), _,_,_).
+sweep_module_description_(_, P, D) :-
+    xref_comment(P, D, _).
+sweep_module_description_(M, _, D) :-
+    atom_concat('sec:', M, S),
+    man_object_property(section(_, _, S, _), summary(D)).
 
 sweep_predicate_references(MFN, Refs) :-
-    term_string(M:F/N, MFN),
-    pi_head(F/N, H),
+    term_string(M:PI, MFN),
+    pi_head(PI, H),
     findall([B, Path, From, Len],
-            (xref_called(Path0, H, B0, _, Line),
+            ((   xref_called(Path0, H, B0, _, Line)
+             ;   xref_called(Path0, M:H, B0, _, Line)
+             ),
              pi_head(B1, B0),
-             term_string(B1, B),
+             (   B1 = M2:F/N
+             ->  true
+             ;   B1 = F/N,
+                 sweep_module_path_(M2, Path0)
+             ),
+             sweep_module_functor_arity_pi_(M2, F, N, B2),
+             term_string(B2, B),
              atom_string(Path0, Path),
              reference_span(Path0, Line, H, From, Len)),
-            Refs,
-            Tail),
-    findall([B, Path, From, Len],
-            (xref_called(Path0, M:H, B0, _, Line),
-             pi_head(B1, B0),
-             term_string(B1, B),
-             atom_string(Path0, Path),
-             reference_span(Path0, Line, H, From, Len)),
-            Tail).
+            Refs).
 
 :- dynamic current_reference_span/2.
 
@@ -359,31 +413,30 @@ reference_span_(Head, goal_term(_, Goal), Beg0, Len) :-
 reference_span_(_, _, _, _) :- true.
 
 sweep_predicate_location(MFN, [Path|Line]) :-
-    term_string(M:F/N, MFN),
+    term_string(M:PI, MFN),
     !,
-    pi_head(F/N, H),
+    pi_head(PI, H),
     (   sweep_predicate_location_(M, H, Path, Line)
     ->  true
     ;   sweep_predicate_location_(H, Path, Line)
     ).
 sweep_predicate_location(FN, [Path|Line]) :-
-    term_string(F/N, FN),
-    !,
-    pi_head(F/N, H),
+    term_string(PI, FN),
+    pi_head(PI, H),
     sweep_predicate_location_(H, Path, Line).
 
 sweep_predicate_apropos(Query0, Matches) :-
     atom_string(Query, Query0),
     findall([S,Path|Line],
-            (prolog_help:apropos(Query, M:F/N, _, P), P >= 0.3,
-             format(string(S), '~W', [M:F/N, [quoted(true), character_escapes(true)]]),
-             pi_head(F/N, Head),
+            (prolog_help:apropos(Query, M:PI, _, P), P >= 0.3,
+             format(string(S), '~W', [M:PI, [quoted(true), character_escapes(true)]]),
+             catch(pi_head(PI, Head), _, fail),
              sweep_predicate_location_(M, Head, Path, Line)),
             Matches, Tail),
     findall([S,Path],
-            (prolog_help:apropos(Query, F/N, _, P), P >= 0.3,
-             format(string(S), '~W', [F/N, [quoted(true), character_escapes(true)]]),
-             pi_head(F/N, Head),
+            (prolog_help:apropos(Query, PI, _, P), P >= 0.3,
+             format(string(S), '~W', [PI, [quoted(true), character_escapes(true)]]),
+             catch(pi_head(PI, Head), _, fail),
              sweep_predicate_location_(Head, Path, Line)),
             Tail).
 
@@ -426,69 +479,71 @@ sweep_predicate_location_(M, H, Path, Line) :-
     predicate_property(M:H, line_count(Line)),
     atom_string(Path0, Path).
 
+sweep_matching_predicates(Bef, Aft, D, M, PIs) :-
+    setof(M:F/A, sweep_matching_predicate(Bef, Aft, D, M, F, A), PIs).
 
-sweep_predicates_collection(Sub, Preds) :-
-    findall(M:F/N,
-            ( current_predicate(M:F/N),
-              pi_head(F/N, H),
-              (   M == system
-              ->  true
-              ;   \+ (predicate_property(M:H, imported_from(M1)), M \= M1)
-              )
-            ),
-            Preds0,
-            Tail0),
-    findall(M:F/N,
-            ( '$autoload':library_index(H, M, _),
-              pi_head(F/N, H)
-            ),
-            Tail0,
-            Tail1),
-    findall(M:F/N,
-            ( xref_defined(SourceId, H, local(_)),
-              (   xref_module(SourceId, M)
-              ->  true
-              ;   M = user
-              ),
-              \+ (predicate_property(M:H, imported_from(M1)), M \= M1),
-              pi_head(F/N, H)
-            ),
-            Tail1,
-            Tail),
-    findall(M:F/N,
-            ( xref_defined(_, H, imported(SourceId)),
-              (   xref_module(SourceId, M)
-              ->  true
-              ;   M = user
-              ),
-              pi_head(F/N, H)
-            ),
-            Tail),
-    list_to_set(Preds0, Preds1),
-    maplist(sweep_predicate_description, Preds1, Preds2),
-    include(sweep_predicate_non_hidden, Preds2, Preds3),
-    (   Sub == []
-    ->  Preds = Preds3
-    ;   include(sweep_predicate_matches(Sub), Preds3, Preds)
+sweep_matching_predicate(Bef, Aft, D, M, F, A) :-
+    sweep_known_predicate(M, F, A),
+    sweep_predicate_matches_(Bef, Aft, F),
+    A >= D.
+
+sweep_predicate_matches_([], Aft, F) :-
+    !,
+    sweep_predicate_matches_aft(Aft, 0, F).
+sweep_predicate_matches_(Bef, Aft, F) :-
+    once(sub_string(F, N, L, _, Bef)),
+    M is N + L,
+    sweep_predicate_matches_aft(Aft, M, F).
+
+sweep_predicate_matches_aft([], _, _) :- !.
+sweep_predicate_matches_aft(A, N, M) :-
+    sub_atom(M, B, _, _, A),
+    B >= N,
+    !.
+
+sweep_known_predicate(M, F, A) :-
+    current_predicate(M:F/A),
+    (   M == system
+    ->  true
+    ;   pi_head(F/A, H),
+        \+ (predicate_property(M:H, imported_from(M1)), M \= M1)
     ).
+sweep_known_predicate(M, F, A) :-
+    '$autoload':library_index(H, M, _),
+    pi_head(F/A, H).
+sweep_known_predicate(M, F, A) :-
+    xref_defined(SourceId, H, How),
+    xref_definition_line(How, _),
+    (   xref_module(SourceId, M)
+    ->  true
+    ;   M = user
+    ),
+    pi_head(F/A, H).
 
-sweep_predicate_matches(Sub, [String|_]) :-
-    sub_string(String, _, _, _, Sub).
+sweep_predicates_collection(S0, Ps) :-
+    (   S0 == []
+    ->  S = ""
+    ;   S = S0
+    ),
+    sweep_matching_predicates([], S, 0, _, PIs),
+    maplist(sweep_format_pi, PIs, Ps).
 
-sweep_predicate_non_hidden([String|_]) :-
-    \+ sub_string(String, _, _, _, ":'$").
-
-sweep_predicate_description(M:F/N, [S|T]) :-
-    sweep_predicate_description_(M, F, N, T),
+sweep_format_pi(M:F/N, [S|T]) :-
+    sweep_module_functor_arity_pi_(M, F, N, MFA),
     format(string(S),
            '~W',
-           [M:F/N, [quoted(true), character_escapes(true)]]).
+           [MFA, [quoted(true), character_escapes(true)]]),
+    (   sweep_predicate_summary(MFA, Summary)
+    ->  atom_string(Summary, T)
+    ;   T = []
+    ).
 
-sweep_predicate_description_(M, F, N, [D]) :-
-    doc_comment(M:F/N, _, D0, _), !, atom_string(D0, D).
-sweep_predicate_description_(_M, F, N, [D]) :-
-    man_object_property(F/N, summary(D0)), !, atom_string(D0, D).
-sweep_predicate_description_(_, _, _, []).
+sweep_predicate_summary(MFA, D) :-
+    doc_comment(MFA, _, D, _).
+sweep_predicate_summary(MFA, D) :-
+    man_object_property(MFA, summary(D)).
+sweep_predicate_summary(_:FA, D) :-
+    man_object_property(FA, summary(D)).
 
 sweep_packs_collection(SearchString, Packs) :-
     prolog_pack:query_pack_server(search(SearchString), true(Packs0), []),
@@ -500,8 +555,9 @@ sweep_pack_info(pack(Name0, _, Desc0, Version0, URLS0), [Name, Desc, Version, UR
     atom_string(Version0, Version),
     maplist(atom_string, URLS0, URLS).
 
-sweep_pack_install(PackName, []) :-
-    atom_string(Pack, PackName), pack_install(Pack, [silent(true), upgrade(true), interactive(false)]).
+sweep_pack_install(PackName, true) :-
+    atom_string(Pack, PackName),
+    pack_install(Pack, [silent(true), upgrade(true), interactive(false)]).
 
 sweep_colourise_query([String|Offset], _) :-
     prolog_colourise_query(String, module(sweep), sweep_handle_fragment_(Offset)).
@@ -550,6 +606,11 @@ sweep_color_normalized_(_, type_error, [Kind0|_], ["type_error"|Kind]) :-
     !,
     Kind0 =.. [Kind1|_],
     atom_string(Kind1, Kind).
+sweep_color_normalized_(_, macro, [String|_], ["macro"|String]) :-
+    !.
+sweep_color_normalized_(_, decl_option, [Opt0|_], ["decl_option"|Opt]) :-
+    !,
+    term_string(Opt0, Opt).
 sweep_color_normalized_(_, Nom0, _, Nom) :-
     atom_string(Nom0, Nom).
 
@@ -620,10 +681,9 @@ sweep_expand_file_name_(Dir, Spec, Exp) :-
                                    solutions(all),
                                    extensions(['', '.pl'])]).
 
-sweep_path_module([], "user")    :- !.
 sweep_path_module(Path0, Module) :-
     atom_string(Path, Path0),
-    xref_module(Path, Module0),
+    sweep_module_path_(Module0, Path),
     atom_string(Module0, Module).
 
 
@@ -785,12 +845,14 @@ sweep_thread_signal([ThreadId|Goal0], _) :-
     term_string(Goal, Goal0),
     thread_signal(ThreadId, Goal).
 
-sweep_local_predicate_export_comment([Path0,F0,A],Comm) :-
+sweep_local_predicate_export_comment([Path0,F0,A,I0],Comm) :-
     atom_string(Path, Path0),
     atom_string(F, F0),
-    doc_comment(_:F/A, Path:_, _Summary, Comment),
+    atom_string(I, I0),
+    compound_name_arguments(PI, I, [F,A]),
+    doc_comment(_:PI, Path:_, _Summary, Comment),
     comment_modes(Comment, Modes),
-    pi_head(F/A, Head),
+    compound_name_arity(Head, F, A),
     member(ModeAndDet, Modes),
     strip_det(ModeAndDet, Head),
     Head =.. [_|Args],
@@ -801,36 +863,40 @@ sweep_local_predicate_export_comment([Path0,F0,A],Comm) :-
                 ]),
     sub_string(Syn0, 6, _, 1, Comm).
 
+strip_det(//(Mode) is _, Mode) :- !.
 strip_det(Mode is _, Mode) :- !.
 strip_det(//(Mode), Mode) :- !.
 strip_det(Mode, Mode).
 
-sweep_functor_arity_pi([M,F0,A|_], PI) :-
-    !, atom_string(F, F0), term_string(M:F/A, PI).
-sweep_functor_arity_pi([F0,A|Path0], PI) :-
+sweep_module_functor_arity_pi([M0,F0,A], PI) :-
+    atom_string(M, M0),
     atom_string(F, F0),
-    pi_head(F/A, Head),
-    sweep_current_module(M0),
-    (   @(predicate_property(M:Head, visible), M0),
-        \+ @(predicate_property(M:Head, imported_from(_)), M0)
-    ->  T = M:F/A
-    ;   xref_defined(_, Head, imported(Other)), xref_module(Other, M)
-    ->  T = M:F/A
-    ;   string(Path0),
-        atom_string(Path, Path0),
-        xref_defined(Path, Head, _)
-    ->  T = M0:F/A
-    ;   T = F/A
-    ),
-    term_string(T, PI).
+    sweep_module_functor_arity_pi_(M, F, A, PI0),
+    term_string(PI0, PI).
+
+sweep_module_functor_arity_pi_(M, F, A, M:F//B) :-
+    sweep_grammar_rule(M, F, A),
+    !,
+    B is A - 2.
+sweep_module_functor_arity_pi_(M, F, A, M:F/A).
+
+sweep_grammar_rule(M, F, A) :-
+    xref_module(Source, M),
+    pi_head(F/A, H),
+    xref_defined(Source, H, dcg).
+sweep_grammar_rule(M, F, A) :-
+    pi_head(M:F/A, H),
+    predicate_property(H, non_terminal).
 
 sweep_current_module(Module) :-
     sweep_main_thread,
     user:sweep_funcall("buffer-file-name", String),
-    string(String),
-    atom_string(Path, String),
-    sweep_module_path_(Module, Path).
-sweep_current_module(user).
+    (   string(String)
+    ->  atom_string(Path, String),
+        sweep_module_path_(Module, Path)
+    ;   Module = user
+    ).
+
 
 sweep_beginning_of_last_predicate(Start, Next) :-
     sweep_source_id(Path),
@@ -861,35 +927,71 @@ sweep_source_id(Path) :-
     string(Path0),
     atom_string(Path, Path0).
 
-sweep_atom_collection(Sub, Col) :-
-    findall(S,
-            (   current_atom(A),
-                atom_string(A, S),
-                sub_string(S, _, _, _, Sub)
-            ),
-            Col).
+sweep_functors_collection([Bef|Aft], Ps) :-
+    setof(F, sweep_matching_functor(Bef, Aft, F), Ps0),
+    maplist(sweep_format_compound, Ps0, Ps).
 
-sweep_predicate_completion_candidates(D, Ps) :-
-    integer(D),
-    sweep_current_module(M),
-    findall(H,
-            (   (   @(predicate_property(H, visible), M)
-                ;   xref_defined(_, H, _)
-                ),
-                pi_head(_/N, H),
-                N - D >= 0
-            ),
-            Hs),
-    maplist(sweep_format_predicate(M, D), Hs, Ps).
+sweep_format_compound(F/A, [S|SP]) :-
+    pi_head(F/A, H),
+    length(NamedArgs, A),
+    maplist(=('$VAR'('_')), NamedArgs),
+    H =.. [F|NamedArgs],
+    term_string(H, S, [quoted(true),
+                       character_escapes(true),
+                       spacing(next_argument),
+                       numbervars(true)]),
+    term_string(_, S, [subterm_positions(SP)]),
+    !.
 
-sweep_format_predicate(M0, D, H0, [S|SP]) :-
-    pi_head(F/N0, H0),
-    N is N0 - D,
+sweep_matching_functor(Bef, Aft, F/A) :-
+    current_functor(F, A),
+    atom(F),
+    term_string(F, S),
+    sweep_matching_atom(Bef, Aft, S).
+
+sweep_compound_functors_collection([Arity,Bef,Aft], Fs) :-
+    setof(F, sweep_matching_functor(Bef, Aft, F/Arity), Fs0),
+    maplist(term_string, Fs0, Fs).
+
+sweep_matching_atom([], Aft, Atom) :-
+    !,
+    sweep_matching_atom_(Aft, 0, Atom).
+sweep_matching_atom(Bef, Aft, Atom) :-
+    once(sub_string(Atom, N, L, _, Bef)),
+    M is N + L,
+    sweep_matching_atom_(Aft, M, Atom).
+
+sweep_matching_atom_([], _, _) :- !.
+sweep_matching_atom_(A, N, M) :-
+    sub_string(M, B, _, _, A),
+    B >= N,
+    !.
+
+sweep_head_functors_collection([Arity,D,M0,Bef,Aft], Fs) :-
+    (   M0 = []
+    ->  true
+    ;   term_string(M, M0)
+    ),
+    sweep_matching_predicates(Bef, Aft, D, M, PIs0),
+    include({Arity}/[_:_/Arity]>>true, PIs0, PIs),
+    maplist([_:Functor/_, Functor]>>true, PIs, Fs0),
+    maplist(term_string, Fs0, Fs).
+
+sweep_heads_collection([D,M0,Bef,Aft], Ps) :-
+    (   M0 = []
+    ->  true
+    ;   term_string(M, M0)
+    ),
+    sweep_matching_predicates(Bef, Aft, D, M, PIs),
+    maplist(sweep_format_head_(D), PIs, Ps).
+
+sweep_format_head_(D, M:F/A, [S|SP]) :-
+    N is A - D,
     length(NamedArgs, N),
     append(NamedArgs, _, OpenNamedArgs),
-    (   @(predicate_property(H0, implementation_module(M)), M0),
-        predicate_argument_names(M:F/N0, As)
-    ->  maplist(name_variable, As, Vs), OpenNamedArgs = Vs
+    (   predicate_argument_names(M:F/A, As, Extra)
+    ->  maplist(name_variable, As, Vs),
+        append(Vs, Extra, OpenNamedArgs)
     ;   maplist(=('$VAR'('_')), NamedArgs)
     ),
     !,
@@ -965,7 +1067,12 @@ sweep_exportable_predicates(Path0, Preds) :-
     findall(D,
             (   xref_defined(Path, D0, _),
                 \+ xref_exported(Path, D0),
-                pi_head(D1, D0),
+                \+ D0 = _:_,
+                pi_head(F/A, D0),
+                (   xref_defined(Path, D0, dcg)
+                ->  B is A - 2, D1 = F//B
+                ;   D1 = F/A
+                ),
                 term_string(D1, D)
             ),
             Preds).
@@ -990,12 +1097,25 @@ sweep_file_path_in_library(Path, Spec) :-
     ;   term_string(Spec1, Spec)
     ).
 
-predicate_argument_names(M:F/A, Args) :-
-    (   M == system
+
+predicate_argument_names(M:F/A, Args, Extra) :-
+    (   sweep_grammar_rule(M, F, A)
+    ->  Extra = [_,_]
+    ;   Extra = []
+    ),
+    sweep_module_functor_arity_pi_(M, F, A, M:PI),
+    (   predicate_argument_names_from_man(M, PI, Args0)
     ->  true
-    ;   sub_atom(M, 0, 1, _, '$')
+    ;   predicate_argument_names_from_pldoc(M, PI, Args0)),
+    arg(2, PI, N),
+    predicate_argument_names_(N, Args0, Args).
+
+
+predicate_argument_names_from_man(M, PI, Args) :-
+    (   pldoc_man:load_man_object(M:PI, _, _, DOM0)
+    ->  true
+    ;   pldoc_man:load_man_object(PI, _, _, DOM0)
     ),
-    pldoc_man:load_man_object(F/A, _, _, DOM0),
     memberchk(element(dt, _, DOM1), DOM0),
     memberchk(element(a, _, DOM2), DOM1),
     catch(findall(Arg,
@@ -1009,17 +1129,16 @@ predicate_argument_names(M:F/A, Args) :-
                       comma_list(CommaSeparatedArgs, ArgsList),
                       member(Arg, ArgsList)
                   ),
-                  Args0),
+                  Args),
           error(syntax_error(_),_),
-          fail),
-    predicate_argument_names_(A, Args0, Args).
-predicate_argument_names(M:F/A, Args) :-
-    doc_comment(M:F/A, _, _, C),
+          fail).
+
+predicate_argument_names_from_pldoc(M, PI, Args) :-
+    doc_comment(M:PI, _, _, C),
     comment_modes(C, ModeAndDets),
     member(ModeAndDet, ModeAndDets),
     strip_det(ModeAndDet, Head),
-    Head =.. [_|Args0],
-    predicate_argument_names_(A, Args0, Args).
+    Head =.. [_|Args].
 
 predicate_argument_names_(Arity, Args0, Args) :-
     length(Args0, Arity),
@@ -1064,11 +1183,10 @@ dep_import(Path, Kind, PI0) -->
     [[Path, PI, Kind]].
 
 
-sweep_format_head([F0|A], R) :-
+sweep_format_head([M0,F0,A,D], R) :-
+    atom_string(M, M0),
     atom_string(F, F0),
-    pi_head(F/A, H),
-    sweep_current_module(M),
-    sweep_format_predicate(M, 0, H, R).
+    sweep_format_head_(D, M:F/A, R).
 
 sweep_format_term([F0,N,P], [S|SP]) :-
     atom_string(F, F0),
@@ -1091,7 +1209,7 @@ sweep_current_functors(A0, Col) :-
     findall([F|A],
             (   current_functor(F0, A),
                 atom(F0),
-                atom_string(F0, F)
+                term_string(F0, F)
             ),
             Col).
 
@@ -1293,3 +1411,14 @@ sweep_breakpoint_range(Id, [Beg|End]) :-
 sweep_breakpoint_file(Id, File) :-
     breakpoint_property(Id, file(File0)),
     atom_string(File0, File).
+
+sweep_expand_macro(String0, String) :-
+    sweep_current_module(M),
+    term_string(Term0, String0, [variable_names(Vs),
+                                 subterm_positions(Pos0),
+                                 module(M)]),
+    functor(Term0, '#', 1),
+    macros:expand_macros(M, Term0, Term, Pos0, _, _, _),
+    term_string(Term, String, [variable_names(Vs), module(M)]).
+
+sweep_variable_start_code(C, _) :- code_type(C, prolog_var_start).
