@@ -43,7 +43,7 @@
 #include <Python.h>
 #include <assert.h>
 
-static atom_t ATOM_None;
+static atom_t ATOM_none;
 static atom_t ATOM_false;
 static atom_t ATOM_true;
 static atom_t ATOM_pydict;
@@ -51,6 +51,9 @@ static atom_t ATOM_tuple;
 static atom_t ATOM_curl;
 static atom_t ATOM_atom;
 static atom_t ATOM_string;
+static atom_t ATOM_file;
+static atom_t ATOM_eval;
+static atom_t ATOM_single;
 
 static functor_t FUNCTOR_python_error3;
 static functor_t FUNCTOR_error2;
@@ -59,9 +62,11 @@ static functor_t FUNCTOR_eq2;
 static functor_t FUNCTOR_hash1;
 static functor_t FUNCTOR_comma2;
 static functor_t FUNCTOR_curl1;
-static functor_t FUNCTOR_tuple2;
+static functor_t FUNCTOR_key_value2;
 static functor_t FUNCTOR_py1;
 static functor_t FUNCTOR_pySet1;
+static functor_t FUNCTOR_prolog1;
+static functor_t FUNCTOR_at1;
 
 static int py_initialize_done = FALSE;
 
@@ -70,6 +75,12 @@ static int py_unify(term_t t, PyObject *obj, int flags);
 static int py_from_prolog(term_t t, PyObject **obj);
 static void py_yield(void);
 static void py_resume(void);
+static PyObject *py_record(term_t t);
+static int py_unify_record(term_t t, PyObject *rec);
+static int py_unify_constant(term_t t, atom_t c);
+static int py_is_record(PyObject *rec);
+static PyObject *py_free_record(PyObject *rec);
+
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 On Windows,  `python3.dll` provides the  stable C API that  relates to
@@ -229,6 +240,7 @@ MyPy_DECREF(PyObject *o)
   }
 }
 
+
 		 /*******************************
 		 *             BLOB             *
 		 *******************************/
@@ -253,17 +265,27 @@ compare_python_object(atom_t a, atom_t b)
 
 static int
 write_python_object(IOSTREAM *s, atom_t symbol, int flags)
-{ PyObject *obj  = PL_blob_data(symbol, NULL, NULL);
-  PyObject *cls = NULL, *name = NULL;
-  const char *str = NULL;
+{ PyObject *obj = PL_blob_data(symbol, NULL, NULL);
 
-  if ( (cls=PyObject_GetAttrString(obj, "__class__")) &&
-       (name=PyObject_GetAttrString(cls, "__name__")) )
-    str = PyUnicode_AsUTF8AndSize(name, NULL);
+  if ( obj )
+  { PyObject *cls = NULL;
+    PyObject *cname = NULL;
+    const char *name;
+
+    if ( (cls=PyObject_GetAttrString(obj, "__class__")) &&
+	 (cname=PyObject_GetAttrString(cls, "__name__")) )
+      name = PyUnicode_AsUTF8AndSize(cname, NULL);
+    else
+      name = "noclass";
+
+    Sfprintf(s, "<py_%Us>(%p)", name, obj);
+
+    Py_CLEAR(cls);
+    Py_CLEAR(cname);
+  } else
+  { Sfprintf(s, "<py_FREED>(0x0)");
+  }
 
-  Sfprintf(s, "<py_%Us>(%p)", str, obj);
-  Py_CLEAR(cls);
-  Py_CLEAR(name);
   return TRUE;
 }
 
@@ -273,7 +295,7 @@ acquire_python_object(atom_t symbol)
   Py_INCREF(obj);
 }
 
-PL_blob_t PY_OBJECT = {
+static PL_blob_t PY_OBJECT = {
   PL_BLOB_MAGIC,
   PL_BLOB_UNIQUE|PL_BLOB_NOCOPY,
   "PyObject",
@@ -289,14 +311,16 @@ unify_py_obj(term_t t, PyObject *obj)
 }
 
 static int
-get_py_obj(term_t t, PyObject **obj, int error)
+get_py_obj(term_t t, PyObject **objp, int error)
 { void *data;
   size_t size;
   PL_blob_t *type;
 
   if ( PL_get_blob(t, &data, &size, &type) && type == &PY_OBJECT )
   { if ( size != 0 )
-    { *(void**)obj = data;
+    { PyObject *obj = data;
+      Py_INCREF(obj);
+      *objp = obj;
       return TRUE;
     } else
       return PL_existence_error("PyObject", t);
@@ -337,7 +361,8 @@ get_py_module(term_t name, PyObject **mod)
       py_module_table = py_new_hashmap();
 
     if ( (obj=py_lookup_hashmap(py_module_table, id)) )
-    { *mod = obj;
+    { Py_INCREF(obj);
+      *mod = obj;
       return TRUE;
     } else
     { PyObject *idobj;
@@ -358,25 +383,30 @@ get_py_module(term_t name, PyObject **mod)
   return FALSE;
 }
 
+
 static PyObject *
 check_error(PyObject *obj)
 { PyObject *ex = PyErr_Occurred();
 
   if ( ex )
-  { PyObject *type, *value, *stack;
+  { PyObject *type = NULL, *tname = NULL, *value = NULL, *stack = NULL;
     term_t t  = PL_new_term_ref();
     term_t av = PL_new_term_refs(3);
 
     PyErr_Fetch(&type, &value, &stack);
-    int rc = ( py_unify(av+0, type, 0) &&
-	       py_unify(av+1, value, 0) &&
-	       (stack ? py_unify(av+2, stack, 0)
-		      : PL_unify_atom(av+2, ATOM_None)) &&
-	       PL_cons_functor_v(t, FUNCTOR_python_error3, av) &&
-	       PL_put_variable(av+0) &&
-	       PL_cons_functor(t, FUNCTOR_error2, t, av+0) &&
-	       PL_raise_exception(t) );
-    (void)rc;
+    if ( (tname=PyObject_GetAttrString(type, "__name__")) &&
+         PL_unify_chars(av+0, PL_ATOM|REP_UTF8, (size_t)-1,
+			PyUnicode_AsUTF8AndSize(tname, NULL)) &&
+	 (value ? py_unify(av+1, value, 0)
+		: py_unify_constant(av+1, ATOM_none)) &&
+	 (stack ? py_unify(av+2, stack, 0)
+		: py_unify_constant(av+2, ATOM_none)) &&
+	 PL_cons_functor_v(t, FUNCTOR_python_error3, av) &&
+	 PL_put_variable(av+0) &&
+	 PL_cons_functor(t, FUNCTOR_error2, t, av+0) )
+      PL_raise_exception(t);
+
+    Py_CLEAR(tname);
     return NULL;
   } else
     return obj;
@@ -397,9 +427,8 @@ py_unify_long(term_t t, PyObject *obj)
     int rc = FALSE;
 
     if ( !hex )
-    { PyObject *builtins = PyEval_GetBuiltins();
+    { PyObject *builtins = PyEval_GetBuiltins(); /* borrowed reference */
       hex = PyDict_GetItemString(builtins , "hex");
-      Py_CLEAR(builtins);
       if ( !hex )
       { term_t ex;
 
@@ -453,7 +482,7 @@ py_unify_unicode(term_t t, PyObject *obj, int flags)
 static int
 py_unify_tuple(term_t t, PyObject *obj, int flags)
 { Py_ssize_t arity = PyTuple_GET_SIZE(obj);
-  if ( PL_unify_functor(t, PL_new_functor(ATOM_tuple, arity)) )
+  if ( PL_unify_compound(t, PL_new_functor(ATOM_tuple, arity)) )
   { term_t a = PL_new_term_ref();
     for(Py_ssize_t i=0; i<arity; i++)
     { PyObject *py_a = PyTuple_GetItem(obj, i);
@@ -605,6 +634,12 @@ out:
 }
 
 
+static int
+py_unify_constant(term_t t, atom_t c)
+{ return PL_unify_term(t, PL_FUNCTOR, FUNCTOR_at1, PL_ATOM, c);
+}
+
+
 static int
 py_unify(term_t t, PyObject *obj, int flags)
 { if ( !obj )
@@ -613,9 +648,9 @@ py_unify(term_t t, PyObject *obj, int flags)
   }
 
   if ( obj == Py_None )
-    return PL_unify_atom(t, ATOM_None);
+    return py_unify_constant(t, ATOM_none);
   if ( PyBool_Check(obj) )
-    return PL_unify_bool(t, PyLong_AsLong(obj));
+    return py_unify_constant(t, PyLong_AsLong(obj) ? ATOM_true : ATOM_false);
 
   if ( (flags&PYU_OBJ) )
   { if ( PyLong_CheckExact(obj) )
@@ -642,6 +677,8 @@ py_unify(term_t t, PyObject *obj, int flags)
       return py_unify_sequence(t, obj, flags);
     if ( PySet_Check(obj) )
       return py_unify_set(t, obj, flags);
+    if ( py_is_record(obj) )
+      return py_unify_record(t, obj);
   }
 
   return unify_py_obj(t, obj);
@@ -684,7 +721,7 @@ py_add_to_dict(term_t key, term_t value, void *closure)
 static int
 add_prolog_key_value_to_dict(PyObject *py_dict, term_t tuple,
 			     term_t key, term_t value)
-{ if ( PL_is_functor(tuple, FUNCTOR_tuple2) )
+{ if ( PL_is_functor(tuple, FUNCTOR_key_value2) )
   { _PL_get_arg(1, tuple, key);
     _PL_get_arg(2, tuple, value);
     PyObject *py_key = NULL, *py_value = NULL;
@@ -719,12 +756,69 @@ py_empty_dict(PyObject **obj)
     return FALSE;
 }
 
+static int
+py_from_prolog_curl(term_t t, PyObject **obj)
+{ term_t iter  = PL_new_term_ref();
+  term_t head  = PL_new_term_ref();
+  term_t key   = PL_new_term_ref();
+  term_t value = head;
+  PyObject *py_dict = PyDict_New();
+
+  _PL_get_arg(1, t, iter);
+  while( PL_is_functor(iter, FUNCTOR_comma2) )
+  { _PL_get_arg(1, iter, head);
+    _PL_get_arg(2, iter, iter);
+
+    if ( !add_prolog_key_value_to_dict(py_dict, head, key, value) )
+      return FALSE;
+  }
+  if ( !add_prolog_key_value_to_dict(py_dict, iter, key, value) )
+    return FALSE;
+
+  PL_reset_term_refs(iter);
+  *obj = py_dict;
+  return TRUE;
+}
+
+
+static int
+py_from_prolog_at1(term_t t, PyObject **obj)
+{ atom_t a;
+  term_t arg = PL_new_term_ref();
+
+  _PL_get_arg(1, t, arg);
+  if ( PL_get_atom(arg, &a) )
+  { int v = -1;
+
+    PL_reset_term_refs(arg);
+
+    if ( a == ATOM_false )
+      v = 0;
+    else if ( a == ATOM_true )
+      v = 1;
+
+    if ( v >= 0 )
+    { *obj = check_error(PyBool_FromLong(v));
+      return !!*obj;
+    }
+
+    if ( a == ATOM_none )
+    { Py_INCREF(Py_None);
+      *obj = Py_None;
+      return TRUE;
+    }
+  }
+
+  return PL_domain_error("py_constant", t);
+}
+
 
 static int
 py_from_prolog(term_t t, PyObject **obj)
 { wchar_t *s;
   size_t len;
   atom_t a;
+  functor_t funct = 0;
 
   // #(Term) --> stringify
   if ( PL_is_functor(t, FUNCTOR_hash1) )
@@ -769,30 +863,6 @@ py_from_prolog(term_t t, PyObject **obj)
     return FALSE;
   }
 
-  // Special atoms
-  if ( PL_get_atom(t, &a) )
-  { int v = -1;
-
-    if ( a == ATOM_false )
-      v = 0;
-    else if ( a == ATOM_true )
-      v = 1;
-
-    if ( v >= 0 )
-    { *obj = check_error(PyBool_FromLong(v));
-      return !!*obj;
-    }
-
-    if ( a == ATOM_None )
-    { Py_INCREF(Py_None);
-      *obj = Py_None;
-      return TRUE;
-    }
-
-    if ( a == ATOM_curl )
-      return py_empty_dict(obj);
-  }
-
   // Normal text representations.  Note that [] does not qualify
   // in SWI-Prolog as an atom
   if ( PL_get_wchars(t, &len, &s, CVT_ATOM|CVT_STRING) )
@@ -818,10 +888,16 @@ py_from_prolog(term_t t, PyObject **obj)
       *obj = list;
     else
       Py_CLEAR(list);
-    return TRUE;
+    return rc;
   }
 
-  if ( PL_is_functor(t, FUNCTOR_pySet1) )
+  int ign = PL_get_functor(t, &funct);
+  (void)ign;
+
+  if ( funct == FUNCTOR_at1 )
+    return py_from_prolog_at1(t, obj);
+
+  if ( funct == FUNCTOR_pySet1 )
   { term_t tail = PL_new_term_ref();
 
     _PL_get_arg(1, t, tail);
@@ -869,43 +945,39 @@ py_from_prolog(term_t t, PyObject **obj)
     *obj = py_dict;
     return TRUE;
   }
-  if ( PL_is_functor(t, FUNCTOR_py1) )
+  if ( funct == FUNCTOR_py1 )
   { term_t a = PL_new_term_ref();
     atom_t c;
+    int rc;
 
     _PL_get_arg(1, t, a);
 
     if ( PL_is_functor(a, FUNCTOR_curl1) )
-    { if ( !PL_put_term(t, a) )
-	return FALSE;
-      PL_reset_term_refs(a);
+    { rc = py_from_prolog_curl(a, obj);
     } else if ( PL_get_atom(a, &c) && c == ATOM_curl )
-    { return py_empty_dict(obj);
+    { rc = py_empty_dict(obj);
     } else
       goto error;
+
+    PL_reset_term_refs(a);
+    return rc;
   }
-  if ( PL_is_functor(t, FUNCTOR_curl1) )
-  { term_t iter  = PL_new_term_ref();
-    term_t head  = PL_new_term_ref();
-    term_t key   = PL_new_term_ref();
-    term_t value = head;
-    PyObject *py_dict = PyDict_New();
+  if ( funct == FUNCTOR_curl1 )
+    return py_from_prolog_curl(t, obj);
 
-    _PL_get_arg(1, t, iter);
-    while( PL_is_functor(iter, FUNCTOR_comma2) )
-    { _PL_get_arg(1, iter, head);
-      _PL_get_arg(2, iter, iter);
-
-      if ( !add_prolog_key_value_to_dict(py_dict, head, key, value) )
-	return FALSE;
+  /* prolog(Term) --> record */
+  if ( funct == FUNCTOR_prolog1 )
+  { term_t a = PL_new_term_ref();
+    PyObject *r;
+    _PL_get_arg(1, t, a);
+    if ( (r=py_record(a)) )
+    { PL_reset_term_refs(a);
+      *obj = r;
+      return TRUE;
     }
-    if ( !add_prolog_key_value_to_dict(py_dict, iter, key, value) )
-      return FALSE;
-
-    PL_reset_term_refs(iter);
-    *obj = py_dict;
-    return TRUE;
+    return FALSE;
   }
+
   /* :(a,b,...) --> Python tuple  */
   if ( PL_get_name_arity(t, &a, &len) && a == ATOM_tuple )
   { PyObject *tp = check_error(PyTuple_New(len));
@@ -937,6 +1009,102 @@ error:
 }
 
 
+		 /*******************************
+		 *           RECORDS            *
+		 *******************************/
+
+static PyObject *
+py_term_constructor(void)
+{ static PyObject *con = NULL;
+
+  if ( !con )
+  { PyObject *janus;
+
+    if ( (janus=mod_janus()) )
+      con = PyObject_GetAttrString(janus, "Term");
+  }
+
+  return con;
+}
+
+static PyObject *
+py_record(term_t t)
+{ record_t rec = PL_record(t);
+
+  if ( rec )
+  { PyObject *r = PyLong_FromLongLong((long long)rec);
+    PyObject *argv = NULL;
+    PyObject *con;
+    PyObject *term = NULL;
+
+    if ( (con=py_term_constructor()) &&
+	 (argv=PyTuple_New(1)) )
+    { Py_INCREF(r);
+      PyTuple_SetItem(argv, 0, r);
+      term = PyObject_CallObject(con, argv);
+    }
+
+    Py_CLEAR(r);
+    Py_CLEAR(argv);
+    return term;
+  }
+
+  Py_SetPrologError(PL_exception(0));
+  return NULL;
+}
+
+
+static int
+py_unify_record(term_t t, PyObject *rec)
+{ PyObject *r = check_error(PyObject_GetAttrString(rec, "_record"));
+  int rc = FALSE;
+
+  if ( r )
+  { term_t tmp;
+    long long v;
+
+    rc = ( (v=PyLong_AsLongLong(r)) &&
+	   (tmp=PL_new_term_ref()) &&
+	   PL_recorded((record_t)v, tmp) &&
+	   PL_unify(t, tmp) );
+    Py_DECREF(r);
+  }
+
+  return rc;
+}
+
+
+static PyObject *
+py_free_record(PyObject *rec)
+{ int rc = FALSE;
+
+  if ( PyLong_Check(rec) )
+  { long long v = PyLong_AsLongLong(rec);
+
+    if ( v )
+      PL_erase((record_t)v);
+    rc = TRUE;
+  }
+
+  if ( rc )
+    Py_RETURN_NONE;
+
+  return NULL;
+}
+
+
+static int
+py_is_record(PyObject *rec)
+{ int rc;
+
+  PyObject *cls = PyObject_GetAttrString(rec, "__class__");
+  rc = (cls == py_term_constructor());
+  Py_CLEAR(cls);
+
+  return rc;
+}
+
+
 		 /*******************************
 		 *	   PROLOG BINDING	*
 		 *******************************/
@@ -1042,9 +1210,7 @@ py_eval(PyObject *obj, term_t func)
   PyObject *py_res = NULL;
 
   if ( !obj && get_py_obj(func, &py_res, FALSE) )
-  { Py_INCREF(py_res);
     return py_res;
-  }
 
   if ( obj && PL_get_chars(func, &attr, CVT_ATOM) )
   { return check_error(PyObject_GetAttrString(obj, attr));
@@ -1058,6 +1224,7 @@ py_eval(PyObject *obj, term_t func)
     } else
     { PyObject *builtins = PyEval_GetBuiltins();
       py_func = PyDict_GetItemString(builtins , PL_atom_chars(fname));
+      /* Both builtins and py_func are borrowed references */
       if ( !py_func )
       { term_t fn;
 
@@ -1065,8 +1232,7 @@ py_eval(PyObject *obj, term_t func)
 	     PL_put_atom(fn, fname) )
 	  PL_existence_error("python_builtin", fn);
       }
-
-      Py_DECREF(builtins);
+      Py_INCREF(py_func);
     }
     if ( !py_func )
       goto out;
@@ -1113,7 +1279,6 @@ py_eval(PyObject *obj, term_t func)
 
 	if ( !py_from_prolog(arg, &py_arg) )
 	  goto out;
-	Py_INCREF(py_arg);
 	PyTuple_SetItem(py_argv, i, py_arg);
       }
     }
@@ -1160,7 +1325,6 @@ unchain(term_t call, PyObject **py_target)
       { rc = PL_type_error("py_target", on);
 	break;
       }
-      Py_INCREF(*py_target);
     } else
     { PyObject *next = py_eval(*py_target, on);
 
@@ -1287,14 +1451,18 @@ py_call3(term_t Call, term_t result, term_t options)
 	  { if ( PyObject_SetAttrString(py_target, attr, py_val) == -1 )
 	      rc = !!check_error(NULL);
 	    if ( rc && result )
-	      rc = PL_unify_atom(result, ATOM_None);
+	      rc = py_unify_constant(result, ATOM_none);
 	  }
 	  Py_CLEAR(py_val);
 	}
       } else
 	rc = PL_domain_error("py_attribute", call);
     } else
-    { rc = !!(py_target = py_eval(py_target, call));
+    { PyObject *py_res = py_eval(py_target, call);
+
+      Py_XDECREF(py_target);
+      py_target = py_res;
+      rc = !!py_target;
       if ( rc && result )
 	rc = py_unify(result, py_target, uflags);
     }
@@ -1329,7 +1497,9 @@ alloc_iter_state(iter_state *state)
 { if ( !state->allocated )
   { iter_state *copy = malloc(sizeof(*state));
     if ( copy )
-      *copy = *state;
+    { *copy = *state;
+      copy->allocated = TRUE;
+    }
     state = copy;
   }
 
@@ -1407,7 +1577,6 @@ py_iter3(term_t Iterator, term_t Result, term_t options, control_t handle)
 	{ py_gil_release(gil_state);
 	  PL_retry_address(alloc_iter_state(state)); /* returns */
 	}
-	free_iter_state(state);
 	rc = !PL_exception(0);
 	goto out;
       }
@@ -1448,8 +1617,16 @@ py_with_gil(term_t goal)
 }
 
 
+static PL_option_t pyrun_options[] =
+{ PL_OPTION("file_name", OPT_STRING),
+  PL_OPTION("start",     OPT_ATOM),
+  PL_OPTIONS_END
+};
+
+
+
 static foreign_t
-py_run(term_t Cmd, term_t Globals, term_t Locals, term_t Result)
+py_run(term_t Cmd, term_t Globals, term_t Locals, term_t Result, term_t options)
 { char *cmd;
 
   if ( PL_get_chars(Cmd, &cmd, CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION) )
@@ -1457,14 +1634,29 @@ py_run(term_t Cmd, term_t Globals, term_t Locals, term_t Result)
     PyObject *result = NULL;
     PyGILState_STATE state;
     int rc;
+    char *file_name = "string";
+    atom_t start = ATOM_file;
+    int start_token;
+
+    if ( !PL_scan_options(options, 0, "py_run_options", pyrun_options,
+			  &file_name, &start) )
+      return FALSE;
+    if ( start == ATOM_file )
+      start_token = Py_file_input;
+    else if ( start == ATOM_eval )
+      start_token = Py_eval_input;
+    else if ( start == ATOM_single )
+      start_token = Py_single_input;
+    else
+      return atom_domain_error("py_run_start", start);
 
     if ( !py_gil_ensure(&state) )
       return FALSE;
 
     if ( (rc = (py_from_prolog(Globals, &globals) &&
 		py_from_prolog(Locals, &locals))) )
-    { PyObject *code = check_error(Py_CompileString(cmd, "string",
-						    Py_file_input));
+    { PyObject *code = check_error(Py_CompileString(cmd, file_name,
+						    start_token));
 
       if ( code )
 	result = check_error(PyEval_EvalCode(code, globals, locals));
@@ -1535,6 +1727,22 @@ py_free(term_t t)
 }
 
 
+static foreign_t
+py_is_object(term_t t)
+{ PyObject *obj = NULL;
+  int rc = get_py_obj(t, &obj, FALSE);
+
+  Py_CLEAR(obj);
+
+  return rc;
+}
+
+
+
+		 /*******************************
+		 *        GIL MANAGEMENT        *
+		 *******************************/
+
 typedef struct
 { PyThreadState *state;
   int		 yielded;
@@ -1574,40 +1782,46 @@ py_resume(void)
 
 install_t
 install_janus(void)
-{ MKATOM(None);
+{ MKATOM(none);
   MKATOM(false);
   MKATOM(true);
   MKATOM(atom);
   MKATOM(string);
-  ATOM_tuple  = PL_new_atom(":");
+  MKATOM(file);
+  MKATOM(eval);
+  MKATOM(single);
+  ATOM_tuple  = PL_new_atom("-");
   ATOM_pydict = PL_new_atom("py");
   ATOM_curl   = PL_new_atom("{}");
 
   MKFUNCTOR(python_error, 3);
   MKFUNCTOR(error, 2);
   MKFUNCTOR(py, 1);
-  FUNCTOR_module2 = PL_new_functor(PL_new_atom(":"), 2);
-  FUNCTOR_eq2     = PL_new_functor(PL_new_atom("="), 2);
-  FUNCTOR_hash1   = PL_new_functor(PL_new_atom("#"), 1);
-  FUNCTOR_comma2  = PL_new_functor(PL_new_atom(","), 2);
-  FUNCTOR_curl1   = PL_new_functor(PL_new_atom("{}"), 1);
-  FUNCTOR_pySet1  = PL_new_functor(PL_new_atom("pySet"), 1);
-  FUNCTOR_tuple2  = PL_new_functor(ATOM_tuple, 2);
+  FUNCTOR_module2    = PL_new_functor(PL_new_atom(":"), 2);
+  FUNCTOR_eq2        = PL_new_functor(PL_new_atom("="), 2);
+  FUNCTOR_hash1      = PL_new_functor(PL_new_atom("#"), 1);
+  FUNCTOR_comma2     = PL_new_functor(PL_new_atom(","), 2);
+  FUNCTOR_curl1      = PL_new_functor(PL_new_atom("{}"), 1);
+  FUNCTOR_pySet1     = PL_new_functor(PL_new_atom("pySet"), 1);
+  FUNCTOR_at1        = PL_new_functor(PL_new_atom("@"), 1);
+  FUNCTOR_key_value2 = FUNCTOR_module2;
+  MKFUNCTOR(prolog, 1);
 
 #define REGISTER(name, arity, func, flags) \
         PL_register_foreign_in_module("janus", name, arity, func, flags)
 
-  REGISTER("py_initialize_", 3, py_initialize_, 0);
-  REGISTER("py_call",        1, py_call1,       0);
-  REGISTER("py_call",        2, py_call2,       0);
-  REGISTER("py_call",        3, py_call3,       0);
-  REGISTER("py_iter",        2, py_iter2,       PL_FA_NONDETERMINISTIC);
-  REGISTER("py_iter",        3, py_iter3,       PL_FA_NONDETERMINISTIC);
-  REGISTER("py_run",         4, py_run,         0);
-  REGISTER("py_free",        1, py_free,        0);
-  REGISTER("py_with_gil",    1, py_with_gil,    PL_FA_TRANSPARENT);
-  REGISTER("py_str",         2, py_str,         0);
-  REGISTER("py_debug",       1, py_debug,       0);
+  REGISTER("py_initialize_",  3, py_initialize_, 0);
+  REGISTER("py_call",         1, py_call1,       0);
+  REGISTER("py_call",         2, py_call2,       0);
+  REGISTER("py_call",         3, py_call3,       0);
+  REGISTER("py_iter",         2, py_iter2,       PL_FA_NONDETERMINISTIC);
+  REGISTER("py_iter",         3, py_iter3,       PL_FA_NONDETERMINISTIC);
+  REGISTER("py_run",          5, py_run,         0);
+  REGISTER("py_free",         1, py_free,        0);
+  REGISTER("py_is_object",    1, py_is_object,   0);
+  REGISTER("py_with_gil",     1, py_with_gil,    PL_FA_TRANSPARENT);
+  REGISTER("py_str",          2, py_str,         0);
+  REGISTER("py_debug",        1, py_debug,       0);
 
   if ( PyImport_AppendInittab("swipl", PyInit_swipl) == -1 )
     Sdprintf("Failed to add module swipl to Python");
