@@ -41,6 +41,7 @@
 	    py_iter/2,			% +Call, -Return
 	    py_iter/3,			% +Call, -Return, +Options
             py_free/1,			% +Obj
+	    py_is_object/1,		% @Term
 	    py_with_gil/1,		% :Goal
 
             pyfunc/3,                   % +Module, +Function, -Return
@@ -61,15 +62,16 @@
             py_pp/2,                    % +Stream, +Term
             py_pp/3,                    % +Stream, +Term, +Options
 
-            obj_dir/2,                  % +ObjRef,-List
-            obj_dict/2,                 % obj_dict(+ObjRef, -Dict)
+            py_obj_dir/2,               % +ObjRef,-List
+            py_obj_dict/2,              % obj_dict(+ObjRef, -Dict)
 
             py_initialize/3,            % +Program, +Argv, +Options
             py_lib_dirs/1,              % -Dirs
-            add_py_lib_dir/1,           % +Dir
-            add_py_lib_dir/2,           % +Dir,+Where
+            py_add_lib_dir/1,           % +Dir
+            py_add_lib_dir/2,           % +Dir,+Where
 
-            op(50,  fx,  #)             % #Value
+            op(200, fy, @),             % @constant
+            op(50,  fx, #)              % #Value
           ]).
 :- use_module(library(apply_macros), []).
 :- autoload(library(lists), [append/3, member/2]).
@@ -96,20 +98,32 @@ add_python_dll_dir :-
 
 :- public
     py_initialize/0,
-    py_call_string/3.
+    py_call_string/3,
+    py_write/2,
+    py_readline/4.
 
 /** <module> Call Python from Prolog
 
-This library implements calling Python from Prolog.
+This library implements calling Python  from   Prolog.  It  is available
+directly from Prolog if the janus   package is bundled, providing access
+to an _embedded_ Python instance. If  SWI-Prolog is embedded into Python
+using the Python package ``janus-swi``, this  library is provided either
+from Prolog or from the Python package.
+
+Normally,  the  Prolog  user  can  simply  start  calling  Python  using
+py_call/2 or friends. In special cases it   may  be needed to initialize
+Python with options using  py_initialize/3   and  optionally  the Python
+search path may be extended using py_add_lib_dir/1.
 */
 
 %!  py_version is det.
 %
-%   Print version info on the available Python installation
+%   Print version info on the  embedded   Python  installation  based on
+%   Python `sys:version`.
 
 py_version :-
     py_call(sys:version, X),
-    print_message(information, py_version(X)).
+    print_message(information, janus(version(X))).
 
 
 %!  py_call(+Call) is det.
@@ -185,10 +199,7 @@ py_version :-
 %     - py_string_as(+Type)
 %       If Type is `atom` (default), translate a Python String into a
 %       Prolog atom.  If Type is `string`, translate into a Prolog string.
-%       Note that by using a string we can distinguish the Python ``Bool``
-%       ``True`` and ``False`` as well as Python ``None`` from the
-%       Python strings ``"true"``, ``"false"`` and ``"None"``.  Strings
-%       are also more efficient if they are short lived.
+%	Strings are more efficient if they are short lived.
 %     - py_object(Boolean)
 %       It `true` (default `false`), translate the return as a Python
 %       object reference unless it is `None`, a boolean or a number.
@@ -235,10 +246,26 @@ py_version :-
 %   @arg Options is processed as with py_call/3.
 %   @bug Iterator may not depend on janus.Query()
 
-%!  py_run(+String, +Globals, +Locals, -Result) is det.
+%!  py_run(+String, +Globals, +Locals, -Result, +Options) is det.
 %
-%   Interface to PyRun_String(). Currently   using  ``Py_file_input`` as
-%   _first token_. Unclear what we can do with this.
+%   Interface  to  Py_CompileString()  followed   by  PyEval_EvalCode().
+%   Options:
+%
+%       - file_name(String)
+%         Errors are reported against this pseudo file name
+%       - start(Token)
+%         One of `eval`, `file` (default) or `single`.
+%
+%   @arg Globals is a dict
+%   @arg Locals is a dict
+
+%!  py_is_object(@Term) is semidet.
+%
+%   True when Term is a Python object reference. Fails silently if @Term
+%   is any other Prolog term.
+%
+%   @error existence_error(py_object, Term) is raised of Term is a
+%   Python object, but it has been freed using py_free/1.
 
 %!  py_free(+Obj) is det.
 %
@@ -425,10 +452,24 @@ comma_dict_items(Key:Value, Keys) =>
 %
 %   If possible, we enable command line   editing using the GNU readline
 %   library.
+%
+%   When used in an environment  where  Prolog   does  not  use the file
+%   handles 0,1,2 for  the  standard   streams,  e.g.,  in  `swipl-win`,
+%   Python's I/O is rebound to use  Prolog's I/O. This includes Prolog's
+%   command line editor, resulting in  a   mixed  history  of Prolog and
+%   Pythin commands.
 
 py_shell :-
-    py_run("from janus import *", _{}, _{}, _),
-    py_call(janus:interact(), _).
+    import_janus,
+    py_call(janus_swi:interact(), _).
+
+import_janus :-
+    py_call(sys:hexversion, V),
+    V >= 0x030A0000,                    % >= 3.10
+    !,
+    py_run("from janus_swi import *", py{}, py{}, _, []).
+import_janus :-
+    print_message(warning, janus(py_shell(no_janus))).
 
 
 		 /*******************************
@@ -469,18 +510,17 @@ opts_kws(Options, Kws) :-
 pair_kws(Name-Value, Name=Value).
 
 
-%!  obj_dir(+ObjRef, -List) is det.
-%!  obj_dict(+ObjRef, -Dict) is det.
+%!  py_obj_dir(+ObjRef, -List) is det.
+%!  py_obj_dict(+ObjRef, -Dict) is det.
 %
-%   Examine attributes of an object. The predicate obj_dir/2 fetches the
-%   names of all attributes,  while  obj_dict/2   gets  a  dict with all
-%   attributes and their values.
+%   Examine attributes of an object.  The predicate py_obj_dir/2 fetches
+%   the names of all attributes, while   py_obj_dict/2  gets a dict with
+%   all attributes and their values.
 
-obj_dir(ObjRef, List) :-
-    py_call(ObjRef:'__dir__'(), List0),
-    maplist(atom_string, List, List0).
+py_obj_dir(ObjRef, List) :-
+    py_call(ObjRef:'__dir__'(), List).
 
-obj_dict(ObjRef, Dict) :-
+py_obj_dict(ObjRef, Dict) :-
     py_call(ObjRef:'__dict__', Dict).
 
 
@@ -509,6 +549,13 @@ py_initialize :-
 %   a no-op.   This predicate is  thread-safe, where the  first thread
 %   initializes Python.
 %
+%   In addition to initializing the Python system, it
+%
+%     - Adds the directory holding `janus.py` to the Python module
+%       search path.
+%     - If Prolog I/O is not connected to the file handles 0,1,2,
+%       it rebinds Python I/O to use the Prolog I/O.
+%
 %   @arg Options is currently ignored.  It will be used to provide
 %   additional configuration options.
 
@@ -517,10 +564,28 @@ py_initialize(Program, Argv, Options) :-
     ->  absolute_file_name(library('python/janus.py'), Janus,
 			   [ access(read) ]),
 	file_directory_name(Janus, PythonDir),
-	add_py_lib_dir(PythonDir, first)
+	py_add_lib_dir(PythonDir, first),
+	py_connect_io
     ;   true
     ).
 
+%!  py_connect_io is det.
+%
+%   If SWI-Prolog console streams are bound to something non-standard,
+%   bind the Python console I/O to our streans.
+
+py_connect_io :-
+    maplist(non_file_stream,
+	    [0-user_input, 1-user_output, 2-user_error],
+	    NonFiles),
+    Call =.. [connect_io|NonFiles],
+    py_call(janus_swi:Call).
+
+non_file_stream(Expect-Stream, Bool) :-
+    (   stream_property(Stream, file_no(Expect))
+    ->  Bool = @false
+    ;   Bool = @true
+    ).
 
 		 /*******************************
 		 *            PATHS		*
@@ -535,20 +600,20 @@ py_lib_dirs(Dirs) :-
     py_call(sys:path, Dirs0),
     maplist(prolog_to_os_filename, Dirs, Dirs0).
 
-%!  add_py_lib_dir(+Dir) is det.
-%!  add_py_lib_dir(+Dir, +Where) is det.
+%!  py_add_lib_dir(+Dir) is det.
+%!  py_add_lib_dir(+Dir, +Where) is det.
 %
 %   Add a directory to the Python  module   search  path.  In the second
-%   form, Where is one of `first`   or `last`. add_py_lib_dir/1 adds the
+%   form, Where is one of `first`   or `last`. py_add_lib_dir/1 adds the
 %   directory as last.
 %
 %   Dir is in Prolog notation. The added   directory  is converted to an
 %   absolute path using the OS notation.
 
-add_py_lib_dir(Dir) :-
-    add_py_lib_dir(Dir, last).
+py_add_lib_dir(Dir) :-
+    py_add_lib_dir(Dir, last).
 
-add_py_lib_dir(Dir, Where) :-
+py_add_lib_dir(Dir, Where) :-
     absolute_file_name(Dir, AbsDir),
     prolog_to_os_filename(AbsDir, OSDir),
     (   Where == last
@@ -598,7 +663,7 @@ py_call_string(M:String, Input, Dict) :-
     VInput = Input,
     (   call(M:Goal)
     *-> bind_status(Status)
-    ;   Status = false,
+    ;   Status = @false,
 	maplist(bind_none, OutVars)
     ).
 
@@ -608,11 +673,11 @@ not_in_projection(Input, Name=Value) :-
     ;   sub_atom(Name, 0, _, _, '_')
     ).
 
-bind_none('None').
+bind_none(@none).
 
 bind_status(Status) :-
     (   '$tbl_delay_list'([])
-    ->  Status = true
+    ->  Status = @true
     ;   Status = "Undefined"
     ).
 
@@ -667,17 +732,17 @@ px_comp(M, P, Tuple, Vars, Set, TV, Ret) :-
 	Goal =.. [Pred|GArgs]
     ;   Goal =.. [Pred|Out]
     ),
-    compound_name_arguments(OTempl0, :, Out),
+    compound_name_arguments(OTempl0, -, Out),
     tv_goal_and_template(TV, Module:Goal, FGoal, OTempl0, OTempl),
     findall(OTempl, FGoal, Ret0),
-    (   Set == true
+    (   Set == @true
     ->  sort(Ret0, Ret)
     ;   Ret = Ret0
     ).
 
-tv_goal_and_template("plain",  Goal, ucall(Goal, TV), Templ, :(Templ,TV)) :- !.
-tv_goal_and_template("delays", Goal, call_delays(Goal, TV), Templ, :(Templ,TV)) :- !.
-tv_goal_and_template("none",   Goal, Goal, Templ, Templ) :- !.
+tv_goal_and_template(plain,  Goal, ucall(Goal, TV), Templ, -(Templ,TV)) :- !.
+tv_goal_and_template(delays, Goal, call_delays(Goal, TV), Templ, -(Templ,TV)) :- !.
+tv_goal_and_template(none,   Goal, Goal, Templ, Templ) :- !.
 tv_goal_and_template(Mode, _, _, _, _) :-
     domain_error("px_comp() truth_vals", Mode).
 
@@ -690,6 +755,55 @@ ucall(Goal, TV) :-
     ;   TV = 2
     ).
 
+		 /*******************************
+		 *          PYTHON I/O          *
+		 *******************************/
+
+%   py_write(+Stream, -String) is det.
+%   py_readline(+Stream, +Size, +Prompt, +Line) is det.
+%
+%   Called from redefined Python console  I/O   to  write/read using the
+%   Prolog streams.
+
+:- '$hide'((py_write/1,
+	    py_readline/4)).
+
+py_write(Stream, String) :-
+    notrace(format(Stream, '~s', [String])).
+
+py_readline(Stream, Size, Prompt, Line) :-
+    notrace(py_readline_(Stream, Size, Prompt, Line)).
+
+py_readline_(Stream, _Size, Prompt, Line) :-
+    prompt1(Prompt),
+    read_line_to_string(Stream, Read),
+    (   Read == end_of_file
+    ->  Line = ""
+    ;   string_concat(Read, "\n", Line),
+	py_add_history(Read)
+    ).
+
+py_add_history(Line) :-
+    ignore(catch(prolog:history(user_input, add(Line)), _, true)).
+
+
+		 /*******************************
+		 *          COMPILING           *
+		 *******************************/
+
+%   py_consult(+File, +Data, +Module) is det.
+%
+%   Support janus.consult(file, data=None, module='user').
+
+:- public py_consult/3.
+py_consult(File, @none, Module) =>
+    consult(Module:File).
+py_consult(File, Data, Module) =>
+    setup_call_cleanup(
+	open_string(Data, In),
+	load_files(Module:File, [stream(In)]),
+	close(In)).
+
 
 		 /*******************************
 		 *           MESSAGES		*
@@ -699,13 +813,16 @@ ucall(Goal, TV) :-
     prolog:error_message//1,
     prolog:message//1.
 
-prolog:error_message(python_error(Type, Value, _Stack)) -->
-    { py_str(Type, PType),
-      py_str(Value, PValue)
+prolog:error_message(python_error(Class, Value, _Stack)) -->
+    { py_str(Value, Message)
     },
-    [ 'Python error ~w:'-[PType], nl,
-      '  ~w'-[PValue]
+    [ 'Python ', ansi(code, "'~w'", [Class]), ':', nl,
+      '  ~w'-[Message]
     ].
 
-prolog:message(py_version(V)) -->
+prolog:message(janus(version(V))) -->
     [ 'Janus embeds Python ~w'-[V] ].
+prolog:message(janus(py_shell(no_janus))) -->
+    [ 'Janus: py_shell/0: Importing janus into the Python shell requires Python 3.10 or later.', nl,
+      'Run "', ansi(code, 'from janus import *', []), '" in the Python shell to import janus.'
+    ].
