@@ -47,11 +47,13 @@
 
 test_janus :-
     run_tests([ janus_data,
+                janus_prolog_data,
                 janus_obj,
                 janus_params,
                 janus_gc,
 		python_call_prolog,
 		janus_iter,
+		janus_errors,
 		xsb_call
               ]).
 
@@ -59,7 +61,7 @@ test_janus :-
     source_file(test_janus, File),
     file_directory_name(File, Dir0),
     directory_file_path(Dir0, tests, Dir),
-    add_py_lib_dir(Dir, first).
+    py_add_lib_dir(Dir, first).
 
 :- begin_tests(janus_data).
 
@@ -86,17 +88,29 @@ test(dict, Z == py{name:bob, age:42}) :-
     py_call(janus:echo(py({name:bob, age:42})), Z).
 test(dict, Z == py{name:"bob", age:42}) :-
     py_call(janus:echo(py{name:bob, age:42}), Z, [py_string_as(string)]).
+test(nodict, Z == py{}) :-
+    py_call(janus:echo(py({})), Z).
+test(nodict, Z == {}) :-
+    py_call(janus:echo({}), Z).
+test(nodict, Z == @true) :-
+    py_call(demo:isstr({}), Z).
+test(bool, Z == @true) :-
+    py_call(janus:echo(@true), Z).
+test(bool, Z == @false) :-
+    py_call(janus:echo(@false), Z).
 test(bool, Z == true) :-
     py_call(janus:echo(true), Z).
 test(bool, Z == false) :-
     py_call(janus:echo(false), Z).
+test(none, Z == @none) :-
+    py_call(janus:echo(@none), Z).
 test(none, Z == 'None') :-
     py_call(janus:echo('None'), Z).
-test(bool, Z == true) :-
+test(bool, Z == "true") :-
     py_call(janus:echo(true), Z, [py_string_as(string)]).
-test(bool, Z == false) :-
+test(bool, Z == "false") :-
     py_call(janus:echo(false), Z, [py_string_as(string)]).
-test(none, Z == 'None') :-
+test(none, Z == "None") :-
     py_call(janus:echo('None'), Z, [py_string_as(string)]).
 test(set, Set == [1, a, false]) :- % True canot be in a Python set??
     py_call(janus:echo(pySet([1,a,false])), pySet(List)),
@@ -127,6 +141,19 @@ test(iterator, L == [1,2,3,4]) :-          % An iterator checks as a sequence
 
 :- end_tests(janus_data).
 
+:- begin_tests(janus_prolog_data).
+
+test(echo, Term =@= Copy) :-
+    Term = f(X,X,_),
+    py_call(janus:echo(prolog(Term)), Copy).
+test(echo, Term =@= Copy) :-
+    Term = f(X,X,_),
+    py_call(janus:echo(prolog(Term)), Ref, [py_object]),
+    py_call(Ref:'__str__'(), Str),
+    term_string(Copy, Str).
+
+:- end_tests(janus_prolog_data).
+
 :- begin_tests(janus_obj).
 
 test(dog, Tricks == ['roll over']) :-
@@ -156,7 +183,7 @@ test(arg, R == py{a:1,b:a,c:x}) :-
 % Therefore we stop this thread and restore the old mode after the
 % test.
 
-test(gc) :-
+test(gc1) :-
     current_prolog_flag(gc_thread, Old),
     set_prolog_gc_thread(false),
     py_call(demo:gced = 0),
@@ -166,7 +193,7 @@ test(gc) :-
     set_prolog_gc_thread(Old),
     py_call(demo:gced, GCed),
     assertion(GCed > 1 000).
-test(gc, [GCed0,GCed] == [10 000, 10 000]) :-
+test(gc2, [GCed0,GCed] == [10 000, 10 000]) :-
     current_prolog_flag(gc_thread, Old),
     set_prolog_gc_thread(false),
     garbage_collect_atoms,
@@ -205,18 +232,20 @@ test(iter, Sum == 10011) :-             % Check we didn't mess up the Prolog sta
 test(undef, Result == py{status:'Undefined'}) :-
     py_call(janus:once(undefined), Result).
 test(russel, List == [py{'X':barber,'Y':barber,status:'Undefined'},
-		      py{'X':barber,'Y':mayor,status:true}]) :-
+		      py{'X':barber,'Y':mayor,status: @true}]) :-
     py_call(demo:shaves(), List0),
     sort('Y', @=<, List0, List).
-test(py_double, Tuples == [1:1,2:1,3:1,
-                           1:2,2:2,3:2,
-                           1:3,2:3,3:3,
-                           1:4,2:4,3:4]) :-
+test(py_double, Tuples == [1-1,2-1,3-1,
+                           1-2,2-2,3-2,
+                           1-3,2-3,3-3,
+                           1-4,2-4,3-4]) :-
     py_call(demo:double_iter(3,4), Tuples).
-test(invalid_nesting, X == true) :-
+test(invalid_nesting, X == @true) :-
     py_call(demo:test_invalid_nesting(), X).
+:- if((py_call(sys:hexversion, V), V >= 0x03080000)).
 test(while, X == [1,2,3]) :-
-    py_call(demo:test_while(), X).
+    py_call(while:test_while(), X).
+:- endif.
 
 :- end_tests(python_call_prolog).
 
@@ -234,12 +263,26 @@ test(double, Pairs == [1-1,1-2,2-1,2-2]) :-
 
 :- end_tests(janus_iter).
 
+:- begin_tests(janus_errors).
+
+test(ex) :-
+    py_call(demo:call_except(foo), Ex),
+    py_call(Ex:'__str__'(), Str),
+    assertion(sub_string(Str,_,_,_,'Unknown procedure: foo/0')).
+
+:- end_tests(janus_errors).
+
 :- begin_tests(xsb_call).
 
-test(reverse, X == :([py{a:py{b:c}}, :(mytuple), 3, 2, 1], 1)) :-
+test(reverse, X == -([py{a:py{b:c}}, -(mytuple), 3, 2, 1], 1)) :-
     py_call(janus:px_qdet(lists, reverse,
-			  [1,2,3,:(mytuple),py{a:py{b:c}}]),
+			  [1,2,3,-(mytuple),py{a:py{b:c}}]),
 	    X).
+test(px_comp1, X == [-(1)-1,-(2)-1]) :-
+    py_call(janus:px_comp(user, between, 1, 2), X).
+test(px_comp2, X == [-(1),-(2)]) :-
+    py_call(janus:px_comp(user, between, 1, 2, truth_vals=none), X).
+
 :- end_tests(xsb_call).
 
 
