@@ -7,9 +7,16 @@ from janus import *
 def int():
     return 42;
 
-def str():
+def hello_world():
     return "Hello World"
 
+def isstr(s):
+    """Boolean test whether s is a string"""
+    if isinstance(s, str):
+        return True
+    else:
+        return False
+
 def dict1():
     return {"a":1, 2:2}
 
@@ -83,14 +90,6 @@ def test_invalid_nesting():
     q1.close()
     return True
 
-def test_while():
-    list=[]
-    q = Query("between(1,3,X)")
-    while ( s := q.next() ):
-        list.append(s['X'])
-    q.close()
-    return list
-
 # Test using generators as custom iterators.
 
 def squares(start, stop):
@@ -135,3 +134,23 @@ class GCAble:
     def __del__(self):
         global gced
         gced += 1
+
+# Test exceptions and terms
+
+def call_except(query, input={}):
+    """Call query, return the exception or none"""
+    try:
+        once(query, input)
+    except Exception as ex:
+        return ex
+    return None
+
+def call_try(n, query, input={}):
+    """Call query `n` times, eating exceptions"""
+    count = 0
+    for i in range(0,n):
+        try:
+            once(query, input)
+        except Exception as ex:
+            count = count+1
+    return count
