@@ -517,7 +517,7 @@ writef(char *fm, ...)
 			} while(0)
 
 static int
-put_string(int (*out)(void*, wint_t), void *closure, String s)
+put_string(int (*out)(void*, wint_t), void *closure, PceString s)
 { int i;
 
   if ( isstrA(s) )
@@ -567,7 +567,7 @@ Comment to avoid mkproto not generating the prototype for this function
 
 static int
 swritefv(int (*out)(void*, wint_t), void *closure,
-	 const String fmt, int argc, const Any argv[])
+	 const PceString fmt, int argc, const Any argv[])
 { int i;
 
   for( i=0; i<fmt->s_size; i++ )
@@ -827,7 +827,7 @@ swritefv(int (*out)(void*, wint_t), void *closure,
 
 static int
 put_void_str(void *ctx, wint_t c)
-{ String s = ctx;
+{ PceString s = ctx;
 
   s->s_size++;
   if ( c > 0xff )
@@ -839,7 +839,7 @@ put_void_str(void *ctx, wint_t c)
 
 static int
 put_str(void *ctx, wint_t c)
-{ String s = ctx;
+{ PceString s = ctx;
 
   str_store(s, s->s_size, c);
   s->s_size++;
@@ -849,7 +849,7 @@ put_str(void *ctx, wint_t c)
 
 
 status
-str_writefv(String s, CharArray format, int argc, const Any *argv)
+str_writefv(PceString s, CharArray format, int argc, const Any *argv)
 { int len;
 
   str_inithdr(s, FALSE);
@@ -888,6 +888,10 @@ extern int vsscanf(const char *, const char *, va_list);
 #define alloca(n) pceMalloc(n)
 #endif /*ALLOCA_BUG*/
 
+#if !(defined(HAVE_VSSCANF) && defined(HAVE_CAST_VA_LIST))
+#define NO_VSSCANF 1
+#endif
+
 Int
 scanstr(char *str, char *fmt, Any *r)
 { int types[SCAN_MAX_ARGS];
@@ -995,11 +999,11 @@ scanstr(char *str, char *fmt, Any *r)
   }
 
   DEBUG(NAME_scan, Cprintf("argn = %d\n", argn));
-#if defined(HAVE_VSSCANF) && defined(HAVE_CAST_VA_LIST)
+#ifndef NO_VSSCANF
   ar = vsscanf(str, fmt, (va_list) ptrs);
 #else
   switch(argn)
-  { case 0:	ar = sscanf(str, fmt); break;
+  { case 0:	ar = 0; break;
     case 1:	ar = sscanf(str, fmt, ptrs[0]); break;
     case 2:	ar = sscanf(str, fmt, ptrs[0], ptrs[1]); break;
     case 3:	ar = sscanf(str, fmt, ptrs[0], ptrs[1], ptrs[2]);
@@ -1265,6 +1269,8 @@ mclock()
 #if !defined(MSLEEP_DONE) && defined(HAVE_NANOSLEEP)
 #define MSLEEP_DONE 1
 #include <errno.h>
+#include <time.h>
+
 
 void
 msleep(int time)
